from typing import Final, Literal, get_args

from pydantic import UUID4, TypeAdapter

from mex.common.models import BaseModel


class LDAPActor(BaseModel):
    """Model class for generic LDAP accounts."""

    objectGUID: UUID4
    sAMAccountName: str | None = None
    mail: list[str] = []


class LDAPPerson(LDAPActor):
    """Model class for LDAP persons."""

    employeeID: str
    givenName: list[str]
    sn: str
    company: str | None = None
    department: str | None = None
    departmentNumber: str | None = None
    displayName: str | None = None
    ou: list[str] = []


class LDAPPersonWithQuery(BaseModel):
    """Wrapper bundling LDAPPerson models with the query string that found them."""

    person: LDAPPerson
    query: str


class LDAPFunctionalAccount(LDAPActor):
    """Model class for LDAP functional accounts."""

    ou: list[Literal["Funktion"]]


AnyLDAPActor = LDAPPerson | LDAPFunctionalAccount
LDAP_MODEL_CLASSES: Final[list[type[AnyLDAPActor]]] = list(get_args(AnyLDAPActor))
AnyLDAPActorsTypeAdapter: TypeAdapter[list[AnyLDAPActor]] = TypeAdapter(
    list[AnyLDAPActor]
)
LDAPFunctionalAccountsTypeAdapter: TypeAdapter[list[LDAPFunctionalAccount]] = (
    TypeAdapter(list[LDAPFunctionalAccount])
)
LDAPPersonsTypeAdapter: TypeAdapter[list[LDAPPerson]] = TypeAdapter(list[LDAPPerson])



from mex.common.models import ExtractedPrimarySource
from mex.common.primary_source.extract import extract_seed_primary_sources
from mex.common.primary_source.transform import (
    transform_seed_primary_source_to_extracted_primary_source,
)


def get_extracted_primary_source_by_name(name: str) -> ExtractedPrimarySource | None:
    """Pick the seed primary source with the given name, transform and return it.

    Args:
        name: Name (`identifierInPrimarySource`) of the primary source

    Returns:
        Extracted primary source if it was found, else None
    """
    for seed_primary_source in extract_seed_primary_sources():
        if seed_primary_source.identifier == name:
            return transform_seed_primary_source_to_extracted_primary_source(
                seed_primary_source
            )
    return None



from typing import Any

from mex.common.connector.http import HTTPConnector
from mex.common.orcid.models import OrcidRecord, OrcidSearchResponse
from mex.common.settings import BaseSettings


class OrcidConnector(HTTPConnector):
    """Connector class for querying Orcid records."""

    def _set_url(self) -> None:
        """Set url of the host."""
        settings = BaseSettings.get()
        self.url = str(settings.orcid_api_url)

    def _check_availability(self) -> None:
        """Send a GET request to verify the host is available."""
        url = f"{self.url.rstrip('/')}/search"
        response = self._send_request("HEAD", url=url, params={})
        response.raise_for_status()

    def build_query(self, filters: dict[str, Any]) -> str:
        """Construct the ORCID API query string."""
        return " AND ".join([f"{key}:{value}" for key, value in filters.items()])

    def get_record_by_id(self, orcid_id: str) -> OrcidRecord:
        """Get a single orcid record by id.

        Args:
            orcid_id: Unique identifier in ORCID system.

        Returns:
            Orcid record of the single matching id.
        """
        endpoint = f"{orcid_id}/record"
        response = self.request(method="GET", endpoint=endpoint)
        return OrcidRecord.model_validate(response)

    def search_records_by_name(  # noqa: PLR0913
        self,
        given_names: str | None = None,
        family_name: str | None = None,
        given_and_family_names: str | None = None,
        filters: dict[str, Any] | None = None,
        skip: int = 0,
        limit: int = 10,
    ) -> OrcidSearchResponse:
        """Search for orcid results for the given filters.

        Args:
            given_names: Optional given name of a person.
            family_name: Optional surname of a person.
            given_and_family_names: Optional full name of a person.
            filters: Key-value pairs representing ORCID search filters.
            skip: How many items to skip for pagination.
            limit: How many items to return in one page.

        Returns:
            Paginated list of orcid results.
        """
        if filters is None:
            filters = {}
        if given_names and (n := given_names.strip()):
            filters["given-names"] = n
        if family_name and (n := family_name.strip()):
            filters["family-name"] = n
        if given_and_family_names and (n := given_and_family_names.strip()):
            filters["given-and-family-names"] = n
        query = self.build_query(filters)
        params = {"q": query or None, "start": str(skip), "rows": str(limit)}
        response = self.request(method="GET", endpoint="search", params=params)
        return OrcidSearchResponse.model_validate(response)



from pydantic.json_schema import (
    GenerateJsonSchema as PydanticJsonSchemaGenerator,
)
from pydantic.json_schema import JsonSchemaValue


class JsonSchemaGenerator(PydanticJsonSchemaGenerator):
    """Customization of the pydantic class for generating JSON schemas."""

    def handle_ref_overrides(self, json_schema: JsonSchemaValue) -> JsonSchemaValue:
        """Disable pydantic behavior to wrap top-level `$ref` keys in an `allOf`.

        For example, pydantic would convert
            {"$ref": "#/$defs/APIType", "examples": ["api-type-1"]}
        into
            {"allOf": {"$ref": "#/$defs/APIType"}, "examples": ["api-type-1"]}
        which is in fact recommended by JSON schema, but we need to disable this
        to stay compatible with mex-editor and mex-model.
        """
        return json_schema



import json
from operator import attrgetter, itemgetter
from typing import Any, cast

import pytest

from mex.common.testing import Joker
from mex.common.types import MergedPrimarySourceIdentifier, Text, TextLanguage
from mex.common.wikidata.models import (
    Aliases,
    Claim,
    Label,
    Labels,
    WikidataOrganization,
)
from mex.common.wikidata.transform import (
    _get_alternative_names,
    _get_clean_short_names,
    get_official_name_label,
    transform_wikidata_organization_to_extracted_organization,
    transform_wikidata_organizations_to_extracted_organizations,
)
from tests.wikidata.conftest import TESTDATA_DIR


def test_transform_wikidata_organization_to_extracted_organization(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> None:
    expected = {
        "identifier": Joker(),
        "hadPrimarySource": Joker(),
        "identifierInPrimarySource": "Q26678",
        "stableTargetId": Joker(),
        "alternativeName": [
            {"value": "alias_en_3", "language": None},
            {"value": "alias_de_2", "language": None},
            {"value": "alias_en_1", "language": None},
            {"value": "alias_en_2", "language": None},
            {"value": "alias_de_1", "language": None},
            {"value": "test_native_name", "language": TextLanguage.DE},
            {"value": "alias_de_3", "language": None},
            {"value": "alias_en_4", "language": None},
        ],
        "geprisId": [],
        "gndId": ["https://d-nb.info/gnd/2005475-0"],
        "isniId": ["https://isni.org/isni/000000012308257X"],
        "officialName": [
            Text(value="TEST", language=TextLanguage.EN),
            Text(value="TEST", language=TextLanguage.DE),
        ],
        "rorId": ["https://ror.org/05vs9tj88", "https://ror.org/044kkbh92"],
        "shortName": [],
        "viafId": ["https://viaf.org/viaf/129013645"],
        "wikidataId": ["http://www.wikidata.org/entity/Q26678"],
    }
    with (TESTDATA_DIR / "items_details.json").open(encoding="utf-8") as fh:
        wikidata_organization = WikidataOrganization.model_validate(json.load(fh)[0])

    extracted_organization = transform_wikidata_organization_to_extracted_organization(
        wikidata_organization, extracted_primary_source_ids["wikidata"]
    )

    assert extracted_organization

    assert sorted(
        extracted_organization.model_dump()["alternativeName"],
        key=itemgetter("value"),
    ) == sorted(cast("list[Any]", expected["alternativeName"]), key=itemgetter("value"))

    assert (
        extracted_organization.model_dump()["identifierInPrimarySource"]
        == expected["identifierInPrimarySource"]
    )
    assert extracted_organization.model_dump()["rorId"] == expected["rorId"]
    assert extracted_organization.model_dump()["wikidataId"] == expected["wikidataId"]
    assert extracted_organization.model_dump()["gndId"] == expected["gndId"]
    assert extracted_organization.model_dump()["isniId"] == expected["isniId"]
    assert extracted_organization.model_dump()["viafId"] == expected["viafId"]


def test_transform_wikidata_organization_to_organization(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> None:
    """Test wikidata organization transformation to extracted organization."""
    expected = {
        "identifier": Joker(),
        "hadPrimarySource": Joker(),
        "identifierInPrimarySource": "Q26678",
        "stableTargetId": Joker(),
        "alternativeName": [
            {"value": "alias_en_3", "language": None},
            {"value": "alias_de_2", "language": None},
            {"value": "alias_en_1", "language": None},
            {"value": "alias_en_2", "language": None},
            {"value": "alias_de_1", "language": None},
            {"value": "test_native_name", "language": TextLanguage.DE},
            {"value": "alias_de_3", "language": None},
            {"value": "alias_en_4", "language": None},
        ],
        "geprisId": [],
        "gndId": ["https://d-nb.info/gnd/2005475-0"],
        "isniId": ["https://isni.org/isni/000000012308257X"],
        "officialName": [
            Text(value="TEST", language=TextLanguage.EN),
            Text(value="TEST", language=TextLanguage.DE),
        ],
        "rorId": ["https://ror.org/05vs9tj88", "https://ror.org/044kkbh92"],
        "shortName": [],
        "viafId": ["https://viaf.org/viaf/129013645"],
        "wikidataId": ["http://www.wikidata.org/entity/Q26678"],
    }

    with (TESTDATA_DIR / "items_details.json").open(encoding="utf-8") as fh:
        wikidata_organizations = [
            WikidataOrganization.model_validate(item) for item in json.load(fh)
        ]

    extracted_organizations = list(
        transform_wikidata_organizations_to_extracted_organizations(
            wikidata_organizations, extracted_primary_source_ids["wikidata"]
        )
    )

    assert len(extracted_organizations) == 1

    extracted_organization_dict = extracted_organizations[0].model_dump()

    assert sorted(
        extracted_organization_dict["alternativeName"],
        key=itemgetter("value"),
    ) == sorted(cast("list[Any]", expected["alternativeName"]), key=itemgetter("value"))

    assert (
        extracted_organization_dict["identifierInPrimarySource"]
        == expected["identifierInPrimarySource"]
    )
    assert extracted_organization_dict["rorId"] == expected["rorId"]
    assert extracted_organization_dict["wikidataId"] == expected["wikidataId"]
    assert extracted_organization_dict["gndId"] == expected["gndId"]
    assert extracted_organization_dict["isniId"] == expected["isniId"]
    assert extracted_organization_dict["viafId"] == expected["viafId"]


def test_get_alternative_names() -> None:
    """Test if all the alternative names are being transformed."""
    raw_aliases = Aliases.model_validate(
        {
            "de": [
                {"language": "de", "value": "alias_de_1"},
                {"language": "de", "value": "alias_de_2"},
                {"language": "de", "value": "alias_de_3"},
            ],
            "en": [
                {"language": "en", "value": "alias_en_1"},
                {"language": "en", "value": "alias_en_2"},
                {"language": "en", "value": "alias_en_3"},
                {"language": "en", "value": "alias_en_4"},
            ],
        }
    )

    raw_native_labels = [
        Claim.model_validate(
            {
                "mainsnak": {
                    "datavalue": {
                        "value": {"language": "de", "text": "test_native_name"}
                    }
                }
            }
        )
    ]
    expected = [
        Text(value="alias_de_1", language=None),
        Text(value="alias_de_2", language=None),
        Text(value="alias_de_3", language=None),
        Text(value="alias_en_1", language=None),
        Text(value="alias_en_2", language=None),
        Text(value="alias_en_3", language=None),
        Text(value="alias_en_4", language=None),
        Text(value="test_native_name", language=TextLanguage.DE),
    ]

    alternative_names = _get_alternative_names(raw_native_labels, raw_aliases)

    assert sorted(alternative_names, key=attrgetter("value")) == sorted(
        expected, key=attrgetter("value")
    )


def test_get_clean_short_names() -> None:
    """Test if clean acronyms are being returned as expected."""
    expected = [
        Text(value="ABCEN", language=TextLanguage.EN),
        Text(value="ABCDE", language=TextLanguage.DE),
    ]
    short_names = [
        {
            "mainsnak": {
                "snaktype": "value",
                "property": "P1813",
                "hash": "6cd9c230521797cef15c529e5bb006a0c51e801e",
                "datavalue": {
                    "value": {"text": "ABCEN", "language": "en"},
                    "type": "monolingualtext",
                },
                "datatype": "monolingualtext",
            },
            "type": "statement",
            "id": "Q679041$AAE01E9A-03EA-424E-A51A-222A4858C4DD",
            "rank": "normal",
        },
        {
            "mainsnak": {
                "snaktype": "value",
                "property": "P1813",
                "hash": "03dcb3e47ca24e8ab90a1b11eb7602ceca2d07ad",
                "datavalue": {
                    "value": {"text": "ABCDE", "language": "de"},
                    "type": "monolingualtext",
                },
                "datatype": "monolingualtext",
            },
        },
    ]

    for short_name in short_names:
        Claim.model_validate(short_name)

    clean_short_names = _get_clean_short_names(
        [Claim.model_validate(acronym) for acronym in short_names]
    )

    assert sorted(clean_short_names, key=attrgetter("value")) == sorted(
        expected, key=attrgetter("value")
    )


@pytest.mark.parametrize(
    ("labels", "expected"),
    [
        (Labels(), None),
        (
            Labels(en=Label(value="Super cool label")),
            Text(value="Super cool label", language=TextLanguage.EN),
        ),
        (
            Labels(
                de=Label(value="Ein toller Bezeichner"),
                en=Label(value="This does not actually matter"),
            ),
            Text(value="Ein toller Bezeichner", language=TextLanguage.DE),
        ),
    ],
)
def test_get_official_name_label(labels: Labels, expected: Text | None) -> None:
    assert get_official_name_label(labels) == expected



from typing import Any

import pytest
from pydantic import ValidationError, computed_field

from mex.common.models import BaseModel


class ComplexDummyModel(BaseModel):
    """Dummy Model with multiple attributes."""

    optional_str: str | None = None
    required_str: str = "default"
    optional_list: list[str] | None = None
    required_list: list[str] = []

    @computed_field(alias="computedInt")  # type: ignore[prop-decorator]
    @property
    def computed_int(self) -> int:
        return 42


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        ({"optional_str": []}, {"optional_str": None}),
        ({"optional_str": [None]}, {"optional_str": None}),
        ({"optional_str": ["value"]}, {"optional_str": "value"}),
        (
            {"required_str": []},
            "Input should be a valid string [type=string_type, input_value=None, "
            "input_type=NoneType]",
        ),
        (
            {"required_str": [None]},
            "Input should be a valid string [type=string_type, input_value=None, "
            "input_type=NoneType]",
        ),
        (
            {"required_str": ["value", "value"]},
            "got multiple values for required_str [type=value_error, "
            "input_value={'required_str': ['value', 'value']}, input_type=dict]",
        ),
        ({"optional_list": None}, {"optional_list": None}),
        ({"optional_list": "value"}, {"optional_list": ["value"]}),
        ({"required_list": None}, {"required_list": []}),
        ({"required_list": "value"}, {"required_list": ["value"]}),
        ({"computed_int": 42}, {"computed_int": 42}),
        ({"computed_int": [42]}, {"computed_int": 42}),
        (
            {"computed_int": 9999999},
            "Cannot set computed fields to custom values! [type=value_error, "
            "input_value={}, input_type=dict]",
        ),
    ],
    ids=[
        "empty list as optional single",
        "None in list as optional single",
        "string in list as optional single",
        "empty list as required single",
        "None in list as required single",
        "strings in list as required single",
        "None as optional list",
        "string as optional list",
        "None as required list",
        "string as required list",
        "correct int as computed single",
        "correct int in list as computed single",
        "false int as computed single",
    ],
)
def test_base_model_listyness_fix(
    data: dict[str, Any], expected: str | dict[str, Any]
) -> None:
    try:
        model = ComplexDummyModel.model_validate(data)
    except Exception as error:  # noqa: BLE001
        assert str(expected) in str(error)  # noqa: PT017
    else:
        actual = model.model_dump()
        assert isinstance(expected, dict)
        for key, value in expected.items():
            assert actual[key] == value


class Pet(BaseModel):
    name: str


class Shelter(Pet):
    inhabitants: list[Pet]


def test_base_model_listyness_fix_only_runs_on_mutable_mapping() -> None:
    # make sure this raises a validation error and does not fail in fix_listyness
    with pytest.raises(
        ValidationError, match="Input should be a valid dictionary or instance of Pet"
    ):
        Shelter(inhabitants="foo")  # type: ignore[call-arg]


class Computer(BaseModel):
    ram: int = 16

    @computed_field  # type: ignore[prop-decorator]
    @property
    def cpus(self) -> int:
        return 42


def test_verify_computed_field_consistency() -> None:
    computer = Computer.model_validate({"cpus": 42})
    assert computer.cpus == 42

    with pytest.raises(
        ValidationError,
        match="Input should be a valid dictionary, validating other types is not "
        "supported for models with computed fields",
    ):
        Computer.model_validate('{"cpus": 1}')

    with pytest.raises(ValidationError, match="Cannot set computed fields"):
        Computer.model_validate({"cpus": 1})

    with pytest.raises(ValidationError, match="Cannot set computed fields"):
        Computer(cpus=99)


def test_field_assignment_on_model_with_computed_field() -> None:
    computer = Computer()

    # computed field cannot be set
    with pytest.raises(
        AttributeError, match="property 'cpus' of 'Computer' object has no setter"
    ):
        computer.cpus = 99  # type: ignore[misc]

    # non-computed field works as expected
    computer.ram = 32


class DummyBaseModel(BaseModel):
    foo: str | None = None


def test_base_model_str() -> None:
    model = DummyBaseModel(foo="bar")
    assert str(model) == "DummyBaseModel(hash='0xf824b9416780b5f')"


def test_base_model_hash() -> None:
    thing1 = DummyBaseModel(foo="4567")
    thing2 = DummyBaseModel(foo="4567")
    thing3 = DummyBaseModel(foo="5678")

    assert hash(thing1) == 2305511607198672667
    assert hash(thing1) == hash(thing2)
    assert hash(thing1) != hash(thing3)



"""Pytest plugin with common MEx-specific fixtures.

Activate by adding `pytest_plugins = ("mex.common.testing.plugin",)`
to the `conftest.py` in your root test folder.
"""

import json
import os
import re
from collections.abc import Generator
from enum import Enum
from pathlib import Path
from typing import Any, cast
from unittest.mock import MagicMock, Mock

import requests
from langdetect import DetectorFactory
from pydantic import AnyUrl
from requests import HTTPError, Response

from mex.common.connector import CONNECTOR_STORE
from mex.common.orcid.connector import OrcidConnector
from mex.common.orcid.models import OrcidRecord, OrcidSearchResponse
from mex.common.primary_source.extract import extract_seed_primary_sources
from mex.common.primary_source.transform import (
    transform_seed_primary_sources_to_extracted_primary_sources,
)
from mex.common.settings import SETTINGS_STORE, BaseSettings
from mex.common.types import MergedPrimarySourceIdentifier
from mex.common.wikidata.connector import WikidataAPIConnector
from mex.common.wikidata.models import WikidataOrganization


class NoOpPytest:
    """No-op pytest drop-in for when dev dependencies are not installed."""

    FixtureRequest = Any
    MonkeyPatch = Any
    fixture = MagicMock()


try:
    import pytest
except ImportError:
    pytest = NoOpPytest  # type: ignore[assignment]


@pytest.fixture(autouse=True)
def patch_reprs(monkeypatch: pytest.MonkeyPatch) -> None:
    """Allow for easier copying of expected output by patching __repr__ methods."""
    monkeypatch.setattr(
        Enum, "__repr__", lambda self: f"{self.__class__.__name__}.{self.name}"
    )
    monkeypatch.setattr(
        AnyUrl, "__repr__", lambda self: f'AnyUrl("{self}", scheme="{self.scheme}")'
    )


@pytest.fixture(autouse=True)
def isolate_assets_dir(
    is_integration_test: bool,  # noqa: FBT001
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Disable the `MEX_ASSETS_DIR` environment variable for unit testing."""
    if not is_integration_test:  # pragma: no cover
        monkeypatch.delenv("MEX_ASSETS_DIR", raising=False)


@pytest.fixture(autouse=True)
def isolate_work_dir(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Set the `MEX_WORK_DIR` environment variable to a temp path for all tests."""
    monkeypatch.setenv("MEX_WORK_DIR", str(tmp_path))


@pytest.fixture(autouse=True)
def settings() -> BaseSettings:
    """Load the settings for this pytest session."""
    return BaseSettings.get()


@pytest.fixture(autouse=True)
def isolate_settings(
    isolate_assets_dir: None,  # noqa: ARG001
    isolate_work_dir: None,  # noqa: ARG001
) -> Generator[None, None, None]:
    """Automatically reset the settings singleton store."""
    SETTINGS_STORE.reset()
    yield
    SETTINGS_STORE.reset()


@pytest.fixture(autouse=True)
def isolate_connectors() -> Generator[None, None, None]:
    """Automatically close all connectors and remove from singleton store."""
    yield
    CONNECTOR_STORE.reset()


@pytest.fixture()
def is_integration_test(request: pytest.FixtureRequest) -> bool:
    """Check the markers of a test to see if this is an integration test."""
    return any(m.name == "integration" for m in request.keywords.get("pytestmark", ()))


@pytest.fixture()
def in_continuous_integration() -> bool:
    """Check the environment variable `CI` to determine whether we are in CI."""
    return os.environ.get("CI") == "true"


@pytest.fixture(autouse=True)
def isolate_langdetect() -> None:
    """Automatically set the language detection seed to a stable value during tests."""
    DetectorFactory.seed = 0


@pytest.fixture(scope="session", autouse=True)
def faker_session_locale() -> list[str]:
    """Configure the default locales used for localizing fake data."""
    return ["de_DE", "en_US"]


@pytest.fixture()
def extracted_primary_source_ids() -> dict[str, MergedPrimarySourceIdentifier]:
    """Return a mapping from `identifierInPrimarySource` to ExtractedPrimarySources."""
    seed_primary_sources = extract_seed_primary_sources()
    extracted_primary_sources = (
        transform_seed_primary_sources_to_extracted_primary_sources(
            seed_primary_sources
        )
    )
    return {
        p.identifierInPrimarySource: p.stableTargetId for p in extracted_primary_sources
    }


@pytest.fixture
def wikidata_organization_raw() -> dict[str, Any]:
    """Return a raw wikidata organization."""
    with (
        Path(__file__).parent / "test_data" / "wikidata_organization_raw.json"
    ).open() as fh:
        return cast("dict[str, Any]", json.load(fh))


@pytest.fixture
def wikidata_organization(
    wikidata_organization_raw: dict[str, Any],
) -> WikidataOrganization:
    """Return a wikidata organization instance."""
    return WikidataOrganization.model_validate(wikidata_organization_raw)


@pytest.fixture
def mocked_wikidata(
    monkeypatch: pytest.MonkeyPatch, wikidata_organization_raw: dict[str, Any]
) -> None:
    """Mock wikidata connector."""
    response_query = Mock(spec=Response, status_code=200)

    session = MagicMock(spec=requests.Session)
    session.get = MagicMock(side_effect=[response_query])

    def mocked_init(self: WikidataAPIConnector) -> None:
        self.session = session

    monkeypatch.setattr(WikidataAPIConnector, "__init__", mocked_init)

    # mock get_wikidata_org_with_org_id

    def get_wikidata_item_details_by_id(
        _self: WikidataAPIConnector, _item_id: str
    ) -> dict[str, str]:
        return wikidata_organization_raw

    monkeypatch.setattr(
        WikidataAPIConnector,
        "get_wikidata_item_details_by_id",
        get_wikidata_item_details_by_id,
    )


@pytest.fixture
def orcid_person_raw() -> dict[str, Any]:
    """Return a raw orcid person."""
    with (Path(__file__).parent / "test_data" / "orcid_person_raw.json").open() as fh:
        return cast("dict[str, Any]", json.load(fh))


@pytest.fixture
def orcid_person_jayne_raw() -> dict[str, Any]:
    """Return a raw orcid person."""
    with (
        Path(__file__).parent / "test_data" / "orcid_person_jayne_raw.json"
    ).open() as fh:
        return cast("dict[str, Any]", json.load(fh))


@pytest.fixture
def orcid_multiple_matches() -> dict[str, Any]:
    """Return a raw orcid person."""
    with (
        Path(__file__).parent / "test_data" / "orcid_multiple_matches.json"
    ).open() as fh:
        return cast("dict[str, Any]", json.load(fh))


@pytest.fixture
def mocked_orcid(
    monkeypatch: pytest.MonkeyPatch,
    orcid_person_raw: dict[str, Any],
    orcid_multiple_matches: dict[str, Any],
    orcid_person_jayne_raw: dict[str, Any],
) -> None:
    """Mock orcid connector."""
    response_query = Mock(spec=Response, status_code=200)

    session = MagicMock(spec=requests.Session)
    session.get = MagicMock(side_effect=[response_query])

    def mocked_init(self: OrcidConnector) -> None:
        self.session = session

    monkeypatch.setattr(OrcidConnector, "__init__", mocked_init)

    def search_records_by_name(  # noqa: PLR0913
        _self: OrcidConnector,
        given_names: str | None = None,
        family_name: str | None = None,  # noqa: ARG001
        given_and_family_names: str | None = None,
        filters: dict[str, Any] | None = None,  # noqa: ARG001
        skip: int = 0,  # noqa: ARG001
        limit: int = 10,  # noqa: ARG001
    ) -> OrcidSearchResponse:
        response = {"result": [], "num-found": 0}
        if given_names == "John":
            response = {"num-found": 1, "result": [orcid_person_raw]}
        elif given_and_family_names == "Jayne Carberry":
            response = {"num-found": 1, "result": [orcid_person_jayne_raw]}
        elif given_names == "Multiple" or given_and_family_names == "Multiple Carberry":
            response = orcid_multiple_matches
        return OrcidSearchResponse.model_validate(response)

    monkeypatch.setattr(
        OrcidConnector, "search_records_by_name", search_records_by_name
    )

    def get_record_by_id(
        _self: OrcidConnector,
        orcid_id: str,
    ) -> OrcidRecord:
        if orcid_id == "0009-0004-3041-5706":
            return OrcidRecord.model_validate(orcid_person_raw)
        if re.match(
            r"^[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}$", orcid_id
        ):
            return OrcidRecord.model_validate(orcid_person_jayne_raw)
        msg = "404 Not Found"
        raise HTTPError(msg)

    monkeypatch.setattr(OrcidConnector, "get_record_by_id", get_record_by_id)



from mex.common.exceptions import MExError, TimedReadTimeout


def test_mex_error_str() -> None:
    assert (
        str(MExError("I", "have", 5, object, "arguments"))
        == "MExError: I, have, 5, <class 'object'>, arguments"
    )


def test_timed_read_timeout_str() -> None:
    assert (
        str(TimedReadTimeout("I have 2", "args", seconds=5.3))
        == "I have 2, args (seconds elapsed=5.300)"
    )



from unittest.mock import MagicMock, Mock

import pytest
from pytest import MonkeyPatch

from mex.common.backend_api.connector import BackendApiConnector


@pytest.fixture
def mocked_backend(monkeypatch: MonkeyPatch) -> MagicMock:
    """Return the mocked request dispatch method of backend connector."""
    mocked_send_request = MagicMock(
        spec=BackendApiConnector._send_request,
        return_value=Mock(json=MagicMock(return_value={})),
        name="mocked_backend_send_request",
    )
    monkeypatch.setattr(BackendApiConnector, "_send_request", mocked_send_request)
    return mocked_send_request



from uuid import UUID

import pytest

from mex.common.models import (
    AdditivePerson,
    ExtractedOrganization,
    ExtractedPerson,
    MergedPerson,
    PersonRuleSetRequest,
    PersonRuleSetResponse,
    PreventivePerson,
    PreviewPerson,
    SubtractivePerson,
)
from mex.common.types import (
    ExtractedPersonIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
)

pytest_plugins = ("mex.common.testing.plugin",)


@pytest.fixture
def extracted_person() -> ExtractedPerson:
    """Return a dummy extracted person for testing purposes."""
    return ExtractedPerson.model_construct(
        identifierInPrimarySource=str(UUID(int=990, version=4)),
        identifier=ExtractedPersonIdentifier.generate(seed=550),
        stableTargetId=MergedPersonIdentifier.generate(seed=876),
        hadPrimarySource=MergedPrimarySourceIdentifier.generate(seed=200),
        affiliation=[MergedOrganizationIdentifier.generate(seed=300)],
        email=["TintzmannM@rki.de"],
        familyName=["Tintzmann"],
        givenName=["Meinrad"],
        fullName=["Meinrad I. Tintzmann"],
        isniId=["https://isni.org/isni/0000000109403744"],
        memberOf=[
            MergedOrganizationalUnitIdentifier.generate(seed=100),
            MergedOrganizationalUnitIdentifier.generate(seed=101),
        ],
        orcidId=["https://orcid.org/0000-0002-9079-593X"],
    )


@pytest.fixture
def merged_person() -> MergedPerson:
    """Return a dummy merged person for testing purposes."""
    return MergedPerson.model_construct(
        identifier=MergedPersonIdentifier.generate(seed=876),
        affiliation=[MergedOrganizationIdentifier.generate(seed=300)],
        email=["TintzmannM@rki.de"],
        familyName=["Tintzmann"],
        givenName=["Meinrad"],
        fullName=["Meinrad I. Tintzmann"],
        isniId=["https://isni.org/isni/0000000109403744"],
        memberOf=[
            MergedOrganizationalUnitIdentifier.generate(seed=100),
            MergedOrganizationalUnitIdentifier.generate(seed=101),
        ],
        orcidId=["https://orcid.org/0000-0002-9079-593X"],
    )


@pytest.fixture
def extracted_organization_rki() -> ExtractedOrganization:
    """Return a dummy rki extracted organization for testing purposes."""
    return ExtractedOrganization.model_construct(
        identifierInPrimarySource="RKI",
        identifier=ExtractedPersonIdentifier.generate(seed=551),
        stableTargetId=MergedOrganizationIdentifier.generate(seed=301),
        hadPrimarySource=MergedPrimarySourceIdentifier.generate(seed=201),
        officialName=[Text(value="Robert Koch-Institut")],
        shortName=[Text(value="RKI")],
        email=["info@rki.de"],
    )


@pytest.fixture
def preview_person() -> PreviewPerson:
    """Return a dummy preview person for testing purposes."""
    return PreviewPerson(
        identifier=MergedPersonIdentifier.generate(seed=876),
        affiliation=[MergedOrganizationIdentifier.generate(seed=300)],
        email=["TintzmannM@rki.de"],
    )


@pytest.fixture
def rule_set_request() -> PersonRuleSetRequest:
    """Return a dummy person rule set request for testing purposes."""
    return PersonRuleSetRequest(
        additive=AdditivePerson(),
        subtractive=SubtractivePerson(fullName="That's not my name!"),
        preventive=PreventivePerson(),
    )


@pytest.fixture
def rule_set_response() -> PersonRuleSetResponse:
    """Return a dummy person rule set response for testing purposes."""
    return PersonRuleSetResponse(
        stableTargetId=MergedPersonIdentifier.generate(seed=876),
        additive=AdditivePerson(),
        subtractive=SubtractivePerson(fullName="That's not my name!"),
        preventive=PreventivePerson(),
    )



from mex.common.fields import (
    ALL_MODEL_CLASSES_BY_NAME,
    FINAL_FIELDS_BY_CLASS_NAME,
    FROZEN_FIELDS_BY_CLASS_NAME,
    INTEGER_FIELDS_BY_CLASS_NAME,
    LINK_FIELDS_BY_CLASS_NAME,
    LITERAL_FIELDS_BY_CLASS_NAME,
    MERGEABLE_FIELDS_BY_CLASS_NAME,
    MUTABLE_FIELDS_BY_CLASS_NAME,
    REFERENCE_FIELDS_BY_CLASS_NAME,
    STRING_FIELDS_BY_CLASS_NAME,
    TEMPORAL_FIELDS_BY_CLASS_NAME,
    TEXT_FIELDS_BY_CLASS_NAME,
    VOCABULARY_FIELDS_BY_CLASS_NAME,
)


def test_all_fields_by_class_names_include_all_classes() -> None:
    assert {
        len(lookup)
        for lookup in (
            FINAL_FIELDS_BY_CLASS_NAME,
            FROZEN_FIELDS_BY_CLASS_NAME,
            INTEGER_FIELDS_BY_CLASS_NAME,
            LINK_FIELDS_BY_CLASS_NAME,
            LITERAL_FIELDS_BY_CLASS_NAME,
            MERGEABLE_FIELDS_BY_CLASS_NAME,
            MUTABLE_FIELDS_BY_CLASS_NAME,
            REFERENCE_FIELDS_BY_CLASS_NAME,
            STRING_FIELDS_BY_CLASS_NAME,
            TEMPORAL_FIELDS_BY_CLASS_NAME,
            TEXT_FIELDS_BY_CLASS_NAME,
            VOCABULARY_FIELDS_BY_CLASS_NAME,
        )
    } == {len(ALL_MODEL_CLASSES_BY_NAME)}



from copy import deepcopy
from datetime import date, datetime, tzinfo
from enum import Enum
from functools import total_ordering
from itertools import zip_longest
from typing import Any, Union, cast, overload

from pandas._libs.tslibs import parsing
from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler, json_schema
from pydantic_core import core_schema
from pytz import timezone


class TemporalEntityPrecision(Enum):
    """Precision levels that are allowed for temporal entity fields."""

    YEAR = "year"
    MONTH = "month"
    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"
    SECOND = "second"
    MICROSECOND = "microsecond"


TEMPORAL_ENTITY_FORMATS_BY_PRECISION = {
    TemporalEntityPrecision.YEAR: r"%Y",
    TemporalEntityPrecision.MONTH: r"%Y-%m",
    TemporalEntityPrecision.DAY: r"%Y-%m-%d",
    TemporalEntityPrecision.HOUR: r"%Y-%m-%dT%H:%M:%SZ",
    TemporalEntityPrecision.MINUTE: r"%Y-%m-%dT%H:%M:%SZ",
    TemporalEntityPrecision.SECOND: r"%Y-%m-%dT%H:%M:%SZ",
    TemporalEntityPrecision.MICROSECOND: r"%Y-%m-%dT%H:%M:%SZ",
}

TEMPORAL_ENTITY_PRECISIONS_BY_ARG_LENGTH = {
    0: TemporalEntityPrecision.YEAR,
    1: TemporalEntityPrecision.YEAR,
    2: TemporalEntityPrecision.MONTH,
    3: TemporalEntityPrecision.DAY,
    4: TemporalEntityPrecision.HOUR,
    5: TemporalEntityPrecision.MINUTE,
    6: TemporalEntityPrecision.SECOND,
    7: TemporalEntityPrecision.MICROSECOND,
}

TIME_PRECISIONS = [
    TemporalEntityPrecision.HOUR,
    TemporalEntityPrecision.MINUTE,
    TemporalEntityPrecision.SECOND,
    TemporalEntityPrecision.MICROSECOND,
]

CET = timezone("CET")  # default assumed timezone
UTC = timezone("UTC")  # required output timezone
YEAR_MONTH_DAY_TIME_REGEX = r"^[0-9]{4}-(?:0[1-9]|1[0-2])-(?:0[1-9]|[12][0-9]|3[01])T(?:[0-1][0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]Z$"  # noqa: E501
YEAR_MONTH_DAY_REGEX = r"^[0-9]{4}-(?:0[1-9]|1[0-2])-(?:0[1-9]|[12][0-9]|3[01])$"
YEAR_MONTH_REGEX = r"^[0-9]{4}-(?:0[1-9]|1[0-2])$"
YEAR_REGEX = r"^[0-9]{4}$"
MAX_DATETIME_ARGUMENTS = 7


@total_ordering
class TemporalEntity:
    """Custom temporal entity with precision detection and timezone normalization."""

    __slots__ = ("date_time", "precision")

    precision: TemporalEntityPrecision
    date_time: datetime
    STR_SCHEMA_PATTERN = r".*"
    ALLOWED_PRECISION_LEVELS = list(TemporalEntityPrecision.__members__.values())
    JSON_SCHEMA_CONFIG: dict[str, str | list[str]] = {
        "examples": [
            "2011",
            "2019-03",
            "2014-08-24",
            "2022-09-30T20:48:35Z",
        ]
    }

    @overload
    def __init__(
        self,
        *args: Union[str, date, datetime, "TemporalEntity"],
        precision: TemporalEntityPrecision | None = None,
        tzinfo: None = None,
    ) -> None: ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *args: int,
        precision: TemporalEntityPrecision | None = None,
        tzinfo: tzinfo | None = None,
    ) -> None: ...  # pragma: no cover

    def __init__(  # noqa: PLR0912
        self,
        *args: Union[int, str, date, datetime, "TemporalEntity"],
        precision: TemporalEntityPrecision | None = None,
        tzinfo: tzinfo | None = None,
    ) -> None:
        """Create a new temporal entity instance.

        Can parse strings, dates, datetimes, other temporal entity objects
        or 0-7 integers for (year, month, day, hour, minute, second, ms).

        For any argument, a precision level is derived.
        When a time component is given, `tzinfo` keyword with timezone information is
        accepted. When no timezone is given, "CET" is assumed.
        If the timezone is not "UTC", the datetime is converted to "UTC".

        Examples:
            TemporalEntity("May 2005")
            TemporalEntity("2002-04-01T23:59")
            TemporalEntity(2009, 1)
            TemporalEntity(date(2009, 9, 30))
            TemporalEntity(datetime(2009, 9, 30, 23, 59, 5, tzinfo=timezone("CET")))
            TemporalEntity(2009, 9, 30, 23, 59, 5, tzinfo=timezone("CET"))
            TemporalEntity(TemporalEntity(2000))
        """
        if len(args) > MAX_DATETIME_ARGUMENTS:
            msg = (
                f"Temporal entity takes at most {MAX_DATETIME_ARGUMENTS} arguments "
                f"({len(args)} given)"
            )
            raise TypeError(msg)

        if len(args) == 1 and isinstance(
            args[0], str | date | datetime | TemporalEntity
        ):
            if tzinfo:
                msg = "Temporal entity does not accept tzinfo in parsing mode"
                raise TypeError(msg)
            if isinstance(args[0], TemporalEntity):
                date_time, parsed_precision = self._parse_temporal_entity(args[0])
            elif isinstance(args[0], datetime):
                date_time, parsed_precision = self._parse_datetime(args[0])
            elif isinstance(args[0], date):
                date_time, parsed_precision = self._parse_date(args[0])
            else:
                date_time, parsed_precision = self._parse_string(args[0])
        elif all(isinstance(a, int) for a in args):
            args = cast("tuple[int, ...]", args)
            date_time, parsed_precision = self._parse_integers(*args, tzinfo=tzinfo)
        else:
            msg = (
                "Temporal entity takes a single str, date, datetime or "
                "TemporalEntity argument or up to 7 integers"
            )
            raise TypeError(msg)

        if precision:
            self._validate_precision(precision)
        else:
            self._validate_precision(parsed_precision)
            precision = parsed_precision

        if precision in TIME_PRECISIONS:
            date_time = date_time.astimezone(UTC)
        else:
            date_time = date_time.replace(tzinfo=None)

        self.date_time = date_time
        self.precision = precision

    @classmethod
    def _validate_precision(cls, precision: TemporalEntityPrecision) -> None:
        """Confirm that the temporal entity can handle the given precision.

        Args:
            precision: a temporal entity precision

        Raises:
            ValueError: If the given precision is not covered by the temporal entity.
        """
        if precision not in cls.ALLOWED_PRECISION_LEVELS:
            allowed_precision_str = str(
                [p.value for p in cls.ALLOWED_PRECISION_LEVELS]
            )[1:-1]
            if len(cls.ALLOWED_PRECISION_LEVELS) == 1:
                error_str = f"Expected precision level to be {allowed_precision_str}"
            else:
                error_str = (
                    f"Expected precision level to be one of {allowed_precision_str}"
                )
            raise ValueError(error_str)

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """Modify the core schema to add validation and serialization rules."""
        return core_schema.json_or_python_schema(
            json_schema=core_schema.chain_schema(
                [
                    core_schema.str_schema(pattern=cls.STR_SCHEMA_PATTERN),
                    core_schema.no_info_plain_validator_function(cls),
                ]
            ),
            python_schema=core_schema.chain_schema(
                [
                    core_schema.is_instance_schema(cls | date | str | TemporalEntity),
                    core_schema.no_info_plain_validator_function(cls),
                ]
            ),
            serialization=core_schema.to_string_ser_schema(when_used="unless-none"),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema_: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> json_schema.JsonSchemaValue:
        """Modify the json schema to add a title, examples and an optional format."""
        json_schema_ = handler(core_schema_)
        json_schema_["title"] = cls.__name__
        json_schema_.update(cls.JSON_SCHEMA_CONFIG)
        return json_schema_

    @staticmethod
    def _parse_integers(
        *args: int, tzinfo: tzinfo | None = None
    ) -> tuple[datetime, TemporalEntityPrecision]:
        """Parse 0-7 integer arguments into a timestamp and deduct the precision."""
        if tzinfo is None:
            tzinfo = CET
        padded = tuple(a or d for a, d in zip_longest(args, (1970, 1, 1, 0, 0, 0, 0)))
        date_time = datetime(*padded, tzinfo=tzinfo)  # type: ignore[arg-type,misc]
        precision = TEMPORAL_ENTITY_PRECISIONS_BY_ARG_LENGTH[len(args)]
        return date_time, precision

    @staticmethod
    def _parse_temporal_entity(
        value: "TemporalEntity",
    ) -> tuple[datetime, TemporalEntityPrecision]:
        """Parse a temporal entity into a new one by copying its attributes."""
        return deepcopy(value.date_time), value.precision

    @staticmethod
    def _parse_string(value: str) -> tuple[datetime, TemporalEntityPrecision]:
        """Parse a string containing a temporal entity using pandas' tslibs."""
        try:
            parsed, precision = parsing.parse_datetime_string_with_reso(  # type: ignore[attr-defined]
                str(value), freq=None, dayfirst=False, yearfirst=True
            )
        except parsing.DateParseError as error:
            raise ValueError(*error.args) from error
        if parsed.tzinfo is None:
            parsed = parsed.replace(tzinfo=CET)
        if precision in ("millisecond", "nanosecond"):
            precision = TemporalEntityPrecision.MICROSECOND.value
        return parsed, TemporalEntityPrecision(precision)

    @staticmethod
    def _parse_datetime(
        value: datetime,
    ) -> tuple[datetime, TemporalEntityPrecision]:
        """Parse a datetime and assume the precision is microseconds."""
        if value.tzinfo is None:
            value = value.replace(tzinfo=CET)
        return deepcopy(value), TemporalEntityPrecision.MICROSECOND

    @staticmethod
    def _parse_date(
        value: date,
    ) -> tuple[datetime, TemporalEntityPrecision]:
        """Parse a date and assume the precision is days."""
        return (
            datetime(value.year, value.month, value.day, tzinfo=CET),
            TemporalEntityPrecision.DAY,
        )

    def __eq__(self, other: object) -> bool:
        """Return whether the given other value is the same as this one."""
        try:
            other_temporal = TemporalEntity(other)  # type: ignore[call-overload]
        except (TypeError, ValueError):
            return False
        return bool(
            self.date_time == other_temporal.date_time
            and self.precision == other_temporal.precision
        )

    def __gt__(self, other: object) -> bool:
        """Return whether the given other value is the greater than this one."""
        try:
            other_temporal = TemporalEntity(other)  # type: ignore[call-overload]
        except TypeError:
            raise NotImplementedError from None
        return bool(self.date_time > other_temporal.date_time)

    def __hash__(self) -> int:
        """Return the hash for this object."""
        return hash((self.date_time, self.precision))

    def __str__(self) -> str:
        """Render temporal entity with format fitting for its precision."""
        return self.date_time.strftime(
            TEMPORAL_ENTITY_FORMATS_BY_PRECISION[self.precision]
        )

    def __repr__(self) -> str:
        """Overwrite the default representation."""
        return f'{self.__class__.__name__}("{self}")'


class Year(TemporalEntity):
    """Parser for temporal entities with year-precision."""

    STR_SCHEMA_PATTERN = YEAR_REGEX
    ALLOWED_PRECISION_LEVELS = [TemporalEntityPrecision.YEAR]
    JSON_SCHEMA_CONFIG = {"examples": ["2024"]}


class YearMonth(TemporalEntity):
    """Parser for temporal entities with month-precision."""

    STR_SCHEMA_PATTERN = YEAR_MONTH_REGEX
    ALLOWED_PRECISION_LEVELS = [TemporalEntityPrecision.MONTH]
    JSON_SCHEMA_CONFIG = {"examples": ["2019-03"]}


class YearMonthDay(TemporalEntity):
    """Parser for temporal entities with day-precision."""

    STR_SCHEMA_PATTERN = YEAR_MONTH_DAY_REGEX
    ALLOWED_PRECISION_LEVELS = [TemporalEntityPrecision.DAY]
    JSON_SCHEMA_CONFIG = {"examples": ["2014-08-24"]}


class YearMonthDayTime(TemporalEntity):
    """Parser for temporal entities with time-precision."""

    STR_SCHEMA_PATTERN = YEAR_MONTH_DAY_TIME_REGEX
    ALLOWED_PRECISION_LEVELS = TIME_PRECISIONS
    JSON_SCHEMA_CONFIG = {"examples": ["2022-09-30T20:48:35Z"]}


TEMPORAL_ENTITY_CLASSES_BY_PRECISION: dict[
    TemporalEntityPrecision, type[Year | YearMonth | YearMonthDay | YearMonthDayTime]
] = {
    TemporalEntityPrecision.YEAR: Year,
    TemporalEntityPrecision.MONTH: YearMonth,
    TemporalEntityPrecision.DAY: YearMonthDay,
    TemporalEntityPrecision.HOUR: YearMonthDayTime,
    TemporalEntityPrecision.MINUTE: YearMonthDayTime,
    TemporalEntityPrecision.SECOND: YearMonthDayTime,
    TemporalEntityPrecision.MICROSECOND: YearMonthDayTime,
}



import json
from unittest.mock import MagicMock, call

import pytest

from mex.common.backend_api.connector import BackendApiConnector
from mex.common.models import (
    AnyExtractedModel,
    AnyPreviewModel,
    ExtractedPerson,
    ItemsContainer,
    MergedPerson,
    PaginatedItemsContainer,
    PersonRuleSetRequest,
    PersonRuleSetResponse,
    PreviewPerson,
)
from mex.common.testing import Joker


@pytest.mark.usefixtures("mocked_backend")
def test_set_authentication_mocked() -> None:
    connector = BackendApiConnector.get()
    assert connector.session.headers["X-API-Key"] == "dummy_write_key"


def test_ingest_mocked(
    mocked_backend: MagicMock, extracted_person: ExtractedPerson
) -> None:
    mocked_return = {"items": [extracted_person]}
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    connector.ingest([extracted_person])

    assert mocked_backend.call_args == call(
        "POST",
        "http://localhost:8080/v0/ingest",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        data=Joker(),
        timeout=10,
    )
    assert (
        json.loads(mocked_backend.call_args.kwargs["data"])
        == ItemsContainer[AnyExtractedModel](items=[extracted_person]).model_dump()
    )


def test_fetch_extracted_items_mocked(
    mocked_backend: MagicMock, extracted_person: ExtractedPerson
) -> None:
    mocked_return = {"items": [extracted_person.model_dump()], "total": 3}
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.fetch_extracted_items(
        query_string="Tintzmann",
        stable_target_id="NGwfzG8ROsrvIiQIVDVy",
        entity_type=["ExtractedPerson", "ExtractedContactPoint"],
        skip=0,
        limit=1,
    )

    assert response.items == [extracted_person]
    assert response.total == 3

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/extracted-item",
        {
            "q": "Tintzmann",
            "stableTargetId": "NGwfzG8ROsrvIiQIVDVy",
            "entityType": ["ExtractedPerson", "ExtractedContactPoint"],
            "referenceField": None,
            "referencedIdentifier": None,
            "skip": "0",
            "limit": "1",
        },
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_get_extracted_item_mocked(
    mocked_backend: MagicMock, extracted_person: ExtractedPerson
) -> None:
    mocked_return = extracted_person.model_dump()
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.get_extracted_item("e3VhxMhEKyjqN5flzLpiEB")

    assert response == extracted_person

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/extracted-item/e3VhxMhEKyjqN5flzLpiEB",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_fetch_merged_items_mocked(
    mocked_backend: MagicMock, merged_person: MergedPerson
) -> None:
    mocked_return = {"items": [merged_person.model_dump()], "total": 3}
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.fetch_merged_items(
        query_string="Tintzmann",
        entity_type=["MergedPerson", "MergedContactPoint"],
        limit=1,
    )

    assert response.items == [merged_person]
    assert response.total == 3

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/merged-item",
        {
            "q": "Tintzmann",
            "identifier": None,
            "entityType": ["MergedPerson", "MergedContactPoint"],
            "referenceField": None,
            "referencedIdentifier": None,
            "skip": "0",
            "limit": "1",
        },
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_fetch_all_merged_items_mocked(
    mocked_backend: MagicMock, merged_person: MergedPerson
) -> None:
    merged_person_json = merged_person.model_dump()
    mocked_backend.return_value.json.side_effect = [
        {"status": "ok"},  # status check
        {"items": [], "total": 103},  # call to gauge the total
        {"items": [merged_person_json] * 100, "total": 103},  # first page 0-99
        {"items": [merged_person_json] * 3, "total": 103},  # second page 100-103
    ]

    connector = BackendApiConnector.get()
    items = list(connector.fetch_all_merged_items(query_string="Tintzmann"))

    # expect 4 calls: status check, get the total, and get two pages 0-99/100-103
    assert len(mocked_backend.call_args_list) == 4

    # expect all 103 persons to be there
    assert items == [merged_person] * 103

    # expect the last call to be made with the correct parameters
    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/merged-item",
        {
            "q": "Tintzmann",
            "identifier": None,
            "entityType": None,
            "referenceField": None,
            "referencedIdentifier": None,
            "skip": "100",
            "limit": "100",
        },
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_get_merged_item_mocked(
    mocked_backend: MagicMock, merged_person: MergedPerson
) -> None:
    mocked_return = merged_person.model_dump()
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.get_merged_item("NGwfzG8ROsrvIiQIVDVy")

    assert response == merged_person

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/merged-item/NGwfzG8ROsrvIiQIVDVy",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_preview_merged_item_mocked(
    mocked_backend: MagicMock,
    preview_person: PreviewPerson,
    rule_set_request: PersonRuleSetRequest,
) -> None:
    mocked_return = preview_person.model_dump()
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.preview_merged_item("NGwfzG8ROsrvIiQIVDVy", rule_set_request)

    assert response == preview_person

    assert mocked_backend.call_args == call(
        "POST",
        "http://localhost:8080/v0/preview-item/NGwfzG8ROsrvIiQIVDVy",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
        data=Joker(),
    )
    assert (
        json.loads(mocked_backend.call_args.kwargs["data"])
        == rule_set_request.model_dump()
    )


def test_fetch_preview_items_mocked(
    mocked_backend: MagicMock,
    preview_person: PreviewPerson,
) -> None:
    preview_response = PaginatedItemsContainer[AnyPreviewModel](
        items=[preview_person], total=92
    )
    mocked_return = preview_response.model_dump()
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.fetch_preview_items(query_string="foobar", limit=1)

    assert response == preview_response

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/preview-item",
        {
            "q": "foobar",
            "identifier": None,
            "entityType": None,
            "referenceField": None,
            "referencedIdentifier": None,
            "skip": "0",
            "limit": "1",
        },
        timeout=10,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
    )


def test_get_rule_set_mocked(
    mocked_backend: MagicMock, rule_set_response: PersonRuleSetResponse
) -> None:
    mocked_backend.return_value.json.return_value = rule_set_response.model_dump()

    connector = BackendApiConnector.get()
    response = connector.get_rule_set("NGwfzG8ROsrvIiQIVDVy")

    assert response == rule_set_response

    assert mocked_backend.call_args == call(
        "GET",
        "http://localhost:8080/v0/rule-set/NGwfzG8ROsrvIiQIVDVy",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )


def test_merged_person_from_login_mocked(
    mocked_backend: MagicMock, merged_person: MergedPerson
) -> None:
    mocked_return = merged_person.model_dump()
    mocked_backend.return_value.json.return_value = mocked_return

    connector = BackendApiConnector.get()
    response = connector.merged_person_from_login()

    assert response == merged_person

    assert mocked_backend.call_args == call(
        "POST",
        "http://localhost:8080/v0/merged-person-from-login",
        None,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
        timeout=10,
    )



from mex.common.models import BaseModel
from mex.common.types import Text


class SeedPrimarySource(BaseModel):
    """Model class for primary sources coming from the raw-data JSON file."""

    identifier: str
    title: list[Text] = []



from mex.common.models import MEX_PRIMARY_SOURCE_STABLE_TARGET_ID
from mex.common.primary_source.extract import extract_seed_primary_sources
from mex.common.primary_source.transform import (
    transform_seed_primary_sources_to_extracted_primary_sources,
)
from mex.common.testing import Joker
from mex.common.types import TextLanguage


def test_transform_seed_primary_sources_to_extracted_primary_sources() -> None:
    extracted_primary_source = (
        transform_seed_primary_sources_to_extracted_primary_sources(
            extract_seed_primary_sources()
        )[0]
    )
    assert extracted_primary_source.model_dump(
        exclude_none=True, exclude_defaults=True
    ) == {
        "hadPrimarySource": MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
        "identifier": Joker(),
        "identifierInPrimarySource": "mex",
        "stableTargetId": MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
        "title": [{"language": TextLanguage.EN, "value": "Metadata Exchange"}],
    }



from mex.common.identity.memory import MemoryIdentityProvider
from mex.common.testing import Joker
from mex.common.types import MergedPrimarySourceIdentifier


def test_get_identifier() -> None:
    assert MemoryIdentityProvider._get_identifier(
        "foo"
    ) == MemoryIdentityProvider._get_identifier("foo")
    assert MemoryIdentityProvider._get_identifier(
        "foo"
    ) != MemoryIdentityProvider._get_identifier("bar")
    identifiers = [
        MemoryIdentityProvider._get_identifier(arg1, arg2)
        for arg1 in ["foo", "bar"]
        for arg2 in ["a", "b", "c"]
    ]
    # make sure each combination of args results in a unique identifier
    assert len(set(identifiers)) == len(identifiers)


def test_assign() -> None:
    provider = MemoryIdentityProvider.get()
    had_primary_source = MergedPrimarySourceIdentifier("00000000000000")
    identifier_in_primary_source = "thing-1"

    new_identity = provider.assign(had_primary_source, identifier_in_primary_source)

    assert new_identity.model_dump() == {
        "hadPrimarySource": had_primary_source,
        "identifierInPrimarySource": identifier_in_primary_source,
        "stableTargetId": Joker(),
        "identifier": Joker(),
    }

    found_identity = provider.assign(had_primary_source, identifier_in_primary_source)

    assert found_identity.model_dump() == {
        "hadPrimarySource": had_primary_source,
        "identifierInPrimarySource": identifier_in_primary_source,
        "stableTargetId": new_identity.stableTargetId,
        "identifier": new_identity.identifier,
    }

    provider.close()
    provider = MemoryIdentityProvider.get()
    fresh_identity = provider.assign(had_primary_source, identifier_in_primary_source)

    assert fresh_identity.model_dump() == {
        "hadPrimarySource": had_primary_source,
        "identifierInPrimarySource": identifier_in_primary_source,
        "stableTargetId": new_identity.stableTargetId,
        "identifier": new_identity.identifier,
    }


def test_fetch_empty() -> None:
    provider = MemoryIdentityProvider.get()
    had_primary_source = MergedPrimarySourceIdentifier("00000000000000")
    identifier_in_primary_source = "thing-1"

    # fetch something that the provider does not know
    identities = provider.fetch(
        had_primary_source=had_primary_source,
        identifier_in_primary_source=identifier_in_primary_source,
    )
    # we get an empty identity list
    assert not identities


def test_fetch_found() -> None:
    provider = MemoryIdentityProvider.get()
    had_primary_source = MergedPrimarySourceIdentifier("00000000000000")
    identifier_in_primary_source = "thing-1"

    # assign this identity first
    identity = provider.assign(had_primary_source, identifier_in_primary_source)

    # fetch the identity we just assigned by stableTargetId
    identities = provider.fetch(
        stable_target_id=identity.stableTargetId,  # type: ignore[arg-type]
    )

    # we get a perfect match
    assert identities == [identity]

    # fetch our identity with identifierInPrimarySource and hadPrimarySource
    identities = provider.fetch(
        had_primary_source=identity.hadPrimarySource,
        identifier_in_primary_source=identity.identifierInPrimarySource,
    )

    # we get a match again
    assert identities == [identity]


def test_metrics() -> None:
    provider = MemoryIdentityProvider.get()
    had_primary_source = MergedPrimarySourceIdentifier("00000000000000")
    assert provider.metrics() == {"database_size": 1}

    provider.assign(had_primary_source, "thing-X")
    assert provider.metrics() == {"database_size": 2}

    provider.assign(had_primary_source, "thing-Y")
    assert provider.metrics() == {"database_size": 3}



from uuid import UUID

import pytest

from mex.common.ldap.models import (
    AnyLDAPActor,
    LDAPFunctionalAccount,
    LDAPPerson,
    LDAPPersonWithQuery,
)
from mex.common.ldap.transform import (
    PersonName,
    analyse_person_string,
    transform_any_ldap_actor_to_extracted_persons_or_contact_points,
    transform_ldap_functional_accounts_to_extracted_contact_points,
    transform_ldap_persons_to_extracted_persons,
    transform_ldap_persons_with_query_to_extracted_persons,
)
from mex.common.models import (
    ExtractedOrganization,
    ExtractedOrganizationalUnit,
)
from mex.common.testing import Joker
from mex.common.types import MergedPrimarySourceIdentifier
from tests.ldap.conftest import SAMPLE_PERSON_ATTRS


@pytest.fixture
def extracted_unit(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> ExtractedOrganizationalUnit:
    return ExtractedOrganizationalUnit(
        name=["MF"],
        hadPrimarySource=extracted_primary_source_ids["ldap"],
        identifierInPrimarySource="mf",
    )


def test_transform_ldap_functional_accounts_to_extracted_contact_points(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> None:
    ldap_functional_account = LDAPFunctionalAccount(
        mail=["mail@example3.com"],
        objectGUID=UUID(int=42, version=4),
        sAMAccountName="samples",
        ou=["Funktion"],
    )

    extracted_contact_points = (
        transform_ldap_functional_accounts_to_extracted_contact_points(
            [ldap_functional_account], extracted_primary_source_ids["ldap"]
        )
    )
    extracted_contact_point = extracted_contact_points[0]

    expected = {
        "email": ["mail@example3.com"],
        "hadPrimarySource": extracted_primary_source_ids["ldap"],
        "identifier": Joker(),
        "identifierInPrimarySource": "00000000-0000-4000-8000-00000000002a",
        "stableTargetId": Joker(),
    }

    assert (
        extracted_contact_point.model_dump(exclude_none=True, exclude_defaults=True)
        == expected
    )


def test_transform_ldap_persons_to_extracted_persons(
    extracted_unit: ExtractedOrganizationalUnit,
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    extracted_organization_rki: ExtractedOrganization,
) -> None:
    ldap_person = LDAPPerson(
        company="RKI",
        department="MF",
        departmentNumber="MF4",
        displayName="Sample, Sam, Dr.",
        employeeID="SampleS",
        givenName="Sam",
        mail=["mail@example2.com"],
        objectGUID=UUID(int=42, version=4),
        ou="MF",
        sAMAccountName="samples",
        sn="Sample",
    )

    extracted_persons = transform_ldap_persons_to_extracted_persons(
        [ldap_person],
        extracted_primary_source_ids["ldap"],
        [extracted_unit],
        extracted_organization_rki,
    )
    extracted_person = extracted_persons[0]

    expected = {
        "email": ["mail@example2.com"],
        "familyName": ["Sample"],
        "fullName": ["Sample, Sam, Dr."],
        "givenName": ["Sam"],
        "hadPrimarySource": str(extracted_primary_source_ids["ldap"]),
        "identifier": Joker(),
        "identifierInPrimarySource": "00000000-0000-4000-8000-00000000002a",
        "affiliation": [str(extracted_organization_rki.stableTargetId)],
        "memberOf": [str(extracted_unit.stableTargetId)],
        "stableTargetId": Joker(),
    }

    assert (
        extracted_person.model_dump(exclude_none=True, exclude_defaults=True)
        == expected
    )


def test_transform_any_ldap_actor_to_extracted_persons_or_contact_points(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    extracted_unit: ExtractedOrganizationalUnit,
    extracted_organization_rki: ExtractedOrganization,
) -> None:
    ldap_actors: list[AnyLDAPActor] = [
        LDAPFunctionalAccount(
            mail=["postfach@example.com"],
            objectGUID=UUID(int=42, version=4),
            sAMAccountName="postfach",
            ou=["Funktion"],
        ),
        LDAPPerson(
            department="MF",
            displayName="Sample, Sam, Dr.",
            employeeID="SampleS",
            givenName="Sam",
            mail=["mail@example.com"],
            objectGUID=UUID(int=43, version=4),
            sAMAccountName="samples",
            sn="Sample",
        ),
    ]

    extracted_actors = transform_any_ldap_actor_to_extracted_persons_or_contact_points(
        ldap_actors,
        [extracted_unit],
        extracted_primary_source_ids["ldap"],
        extracted_organization_rki,
    )

    assert [
        a.model_dump(exclude_none=True, exclude_defaults=True) for a in extracted_actors
    ] == [
        {
            "hadPrimarySource": str(extracted_primary_source_ids["ldap"]),
            "identifierInPrimarySource": "00000000-0000-4000-8000-00000000002a",
            "email": ["postfach@example.com"],
            "identifier": Joker(),
            "stableTargetId": Joker(),
        },
        {
            "hadPrimarySource": extracted_primary_source_ids["ldap"],
            "identifierInPrimarySource": "00000000-0000-4000-8000-00000000002b",
            "email": ["mail@example.com"],
            "familyName": ["Sample"],
            "fullName": ["Sample, Sam, Dr."],
            "givenName": ["Sam"],
            "memberOf": [str(extracted_unit.stableTargetId)],
            "affiliation": [str(extracted_organization_rki.stableTargetId)],
            "identifier": Joker(),
            "stableTargetId": Joker(),
        },
    ]


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", []),
        ("-", []),
        ("Sur", [PersonName("Sur", "*", "Sur")]),
        ("Given Sur", [PersonName("Sur", "Given", "Given Sur")]),
        ("Given M. Sur-Name", [PersonName("Sur-Name", "Given", "Given Sur-Name")]),
        ("Sur, Given", [PersonName("Sur", "Given", "Given Sur")]),
        (
            "Olaf One, Tracy Two, Toni Three",
            [
                PersonName("One", "Olaf", "Olaf One"),
                PersonName("Two", "Tracy", "Tracy Two"),
                PersonName("Three", "Toni", "Toni Three"),
            ],
        ),
        (
            "Olaf One, Tracy Two",
            [
                PersonName("One", "Olaf", "Olaf One"),
                PersonName("Two", "Tracy", "Tracy Two"),
            ],
        ),
        (
            "Olga One / Troy Two",
            [
                PersonName("One", "Olga", "Olga One"),
                PersonName("Two", "Troy", "Troy Two"),
            ],
        ),
        (
            "Lastname; Given Surname",
            [
                PersonName("Lastname", "*", "Lastname"),
                PersonName("Surname", "Given", "Given Surname"),
            ],
        ),
        (
            "Muster, Max; Doe, Jane",
            [
                PersonName("Muster", "Max", "Max Muster"),
                PersonName("Doe", "Jane", "Jane Doe"),
            ],
        ),
        (
            "Mac Fly, Marty",
            [PersonName("Mac Fly", "Marty", "Marty Mac Fly")],
        ),
        (
            "Multiple Given Names Surname",
            [
                PersonName(
                    "Surname", "Multiple Given Names", "Multiple Given Names Surname"
                )
            ],
        ),
        ("Sur, Given (Leitung Unit)", [PersonName("Sur", "Given", "Given Sur")]),
        ("Sur, Given (FG99", [PersonName("Sur", "Given", "Given Sur")]),
        (
            "Prof. Dr. med. Sur stellv. Projektleitung ABC2",
            [PersonName("Sur", "*", "Sur")],
        ),
        ("Given Sur, Abteilung X3 Z.", [PersonName("Sur", "Given", "Given Sur")]),
        ("Fr. Given Sur 3", [PersonName("Sur", "Given", "Given Sur")]),
    ],
)
def test_analyse_person_string(string: str, expected: list[PersonName]) -> None:
    names = analyse_person_string(string)
    assert names == expected


def test_transform_ldap_persons_with_query_to_extracted_persons(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    extracted_unit: ExtractedOrganizationalUnit,
    extracted_organization_rki: ExtractedOrganization,
) -> None:
    ldap_person = LDAPPerson.model_validate(SAMPLE_PERSON_ATTRS)
    ldap_persons_with_query = [LDAPPersonWithQuery(person=ldap_person, query="test")]

    extracted_persons = transform_ldap_persons_with_query_to_extracted_persons(
        ldap_persons_with_query,
        extracted_primary_source_ids["ldap"],
        [extracted_unit],
        extracted_organization_rki,
    )

    assert len(extracted_persons) == 1
    assert extracted_persons[0].model_dump() == {
        "hadPrimarySource": extracted_primary_source_ids["ldap"],
        "identifierInPrimarySource": "00000000-0000-4000-8000-000000000000",
        "affiliation": [extracted_organization_rki.stableTargetId],
        "email": ["SampleS@mail.tld"],
        "familyName": ["Sample"],
        "fullName": ["Sample, Sam"],
        "givenName": ["Sam"],
        "isniId": [],
        "memberOf": [],
        "orcidId": [],
        "entityType": "ExtractedPerson",
        "identifier": "cXTehc7a4YxNw6j7UpSq0b",
        "stableTargetId": "fWYZKuwfymRj0ItP1CfTO4",
    }



from enum import StrEnum
from typing import Annotated, Any

from pydantic import BaseModel, Field, model_validator

URL_PATTERN = (
    "^(?:(?:[^:/?#]+):)?(?://(?:[^/?#]*))?(?:[^?#]*)(?:\\?(?:[^#]*))?(?:#(?:.*))?$"
)

UrlStr = Annotated[
    str,
    Field(
        pattern=URL_PATTERN,
        min_length=1,
        examples=["https://hello-world.org", "file://S:/OE/MF4/Projekte/MEx"],
        json_schema_extra={"format": "uri"},
    ),
]


class LinkLanguage(StrEnum):
    """Possible language tags for `Link` values."""

    DE = "de"
    EN = "en"
    FR = "fr"
    ES = "es"
    RU = "ru"


class Link(BaseModel):
    """Type class for Link objects.

    Links can be parsed from nested JSON objects or from raw strings.

    Example:
        Link(url="http://foo.bar") == Link.model_validate("http://foo.bar")
    """

    language: LinkLanguage | None = None
    title: str | None = None
    url: UrlStr

    @model_validator(mode="before")
    @classmethod
    def validate_strings(cls, value: Any) -> dict[str, Any]:  # noqa: ANN401
        """Convert string input to dictionary."""
        if isinstance(value, str):
            return {"url": value}
        if isinstance(value, dict):
            return value
        msg = f"Allowed input types are dict and str, got {type(value)}"
        raise ValueError(msg)

    def __hash__(self) -> int:
        """Return the hash of this link."""
        return hash((self.url, self.title, self.language))



import pytest

from mex.common.models import (
    ExtractedOrganization,
    ExtractedOrganizationalUnit,
)
from mex.common.organigram.models import OrganigramUnit
from mex.common.organigram.transform import (
    transform_organigram_units_to_organizational_units,
)
from mex.common.types import Link, LinkLanguage, MergedPrimarySourceIdentifier, Text


@pytest.fixture
def child_unit() -> OrganigramUnit:
    """Return a child unit corresponding to the test_data."""
    return OrganigramUnit(
        shortName=[Text(value="C1")],
        alternativeName=[
            Text(value="CHLD"),
            Text(value="C1 Sub-Unit"),
            Text(value="C1 Unterabteilung"),
        ],
        identifier="child-unit",
        name=[
            Text(value="CHLD Unterabteilung", language="de"),
            Text(value="C1: Sub Unit", language="en"),
        ],
        parentUnit="parent-unit",
    )


@pytest.fixture
def extracted_child_unit(
    child_unit: OrganigramUnit,
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    rki_organization: ExtractedOrganization,
) -> ExtractedOrganizationalUnit:
    """Return the child unit transformed to an ExtractedOrganizationalUnit."""
    return transform_organigram_units_to_organizational_units(
        [child_unit], extracted_primary_source_ids["organigram"], rki_organization
    )[0]


@pytest.fixture
def parent_unit() -> OrganigramUnit:
    """Return a parent unit corresponding to the test_data."""
    return OrganigramUnit(
        shortName=[Text(value="PRNT")],
        alternativeName=[Text(value="PRNT Abteilung"), Text(value="PARENT Dept.")],
        identifier="parent-unit",
        name=[
            Text(value="Abteilung", language="de"),
            Text(value="Department", language="en"),
        ],
        email=["pu@example.com", "PARENT@example.com"],
        website=Link(
            language=LinkLanguage.EN,
            title="Example | Parent Department",
            url="https://www.example.com/departments/parent.html",
        ),
    )


@pytest.fixture
def extracted_parent_unit(
    parent_unit: OrganigramUnit,
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    rki_organization: ExtractedOrganization,
) -> ExtractedOrganizationalUnit:
    """Return the parent unit transformed to an ExtractedOrganizationalUnit."""
    return transform_organigram_units_to_organizational_units(
        [parent_unit], extracted_primary_source_ids["organigram"], rki_organization
    )[0]


@pytest.fixture
def rki_organization() -> ExtractedOrganization:
    """Return a mock RKI organization for testing."""
    return ExtractedOrganization(
        officialName=[Text(value="Robert Koch-Institut")],
        hadPrimarySource=MergedPrimarySourceIdentifier("RKIID1234567890"),
        identifierInPrimarySource="RKIID9876543210",
    )



from pydantic import BaseModel


class Status(BaseModel):
    """Model for system status responses."""

    status: str


class VersionStatus(Status):
    """Model for system status responses with a version."""

    version: str



import pytest
from requests import HTTPError

from mex.common.exceptions import EmptySearchResultError, FoundMoreThanOneError
from mex.common.orcid.extract import (
    get_orcid_record_by_id,
    get_orcid_record_by_name,
    search_records_by_name,
)
from mex.common.orcid.models import (
    OrcidEmails,
    OrcidFamilyName,
    OrcidGivenNames,
    OrcidIdentifier,
    OrcidName,
    OrcidPerson,
    OrcidRecord,
)


@pytest.mark.parametrize(
    ("given_names", "family_name", "given_and_familyname", "expected_result", "error"),
    [
        (
            "John",
            "Doe",
            None,
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[]),
                    name=OrcidName(
                        family_name=OrcidFamilyName(value="Doe"),
                        given_names=OrcidGivenNames(value="John"),
                        visibility="public",
                    ),
                ),
            ),
            None,
        ),
        (
            None,
            None,
            "Jayne Carberry",
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0000-0003-4634-4047",
                    uri="https://orcid.org/0000-0003-4634-4047",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[]),
                    name=OrcidName(
                        family_name=OrcidFamilyName(value="Carberry"),
                        given_names=OrcidGivenNames(value="Jayne"),
                        visibility="public",
                    ),
                ),
            ),
            None,
        ),
        (
            "NotExistJohn",
            "Doe",
            None,
            None,
            EmptySearchResultError,
        ),
        (
            "Multiple",
            "Doe",
            None,
            None,
            FoundMoreThanOneError,
        ),
        (
            None,
            None,
            "Multiple Carberry",
            None,
            FoundMoreThanOneError,
        ),
    ],
    ids=["existing", "existing_jayne", "not_existing", "multiple", "multiple_jayne"],
)
@pytest.mark.usefixtures("mocked_orcid")
def test_get_orcid_record_by_name(
    given_names: str,
    family_name: str,
    given_and_familyname: str,
    expected_result: OrcidRecord,
    error: type[Exception],
) -> None:
    if error:
        with pytest.raises(error):
            get_orcid_record_by_name(
                given_names=given_names,
                family_name=family_name,
                given_and_family_names=given_and_familyname,
            )
    else:
        result = get_orcid_record_by_name(
            given_names=given_names,
            family_name=family_name,
            given_and_family_names=given_and_familyname,
        )
        assert result == expected_result


@pytest.mark.usefixtures("mocked_orcid")
@pytest.mark.parametrize(
    ("orcid_id", "expected_result", "error", "error_message"),
    [
        (
            "0009-0004-3041-5706",
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[]),
                    name=OrcidName(
                        family_name=OrcidFamilyName(value="Doe"),
                        given_names=OrcidGivenNames(value="John"),
                        visibility="public",
                    ),
                ),
            ),
            None,
            None,
        ),
        ("0000-0000-0000-000", None, HTTPError, "404 Not Found"),
    ],
    ids=["existing id", "non-existing id"],
)
def test_get_orcid_record_by_id(
    orcid_id: str,
    expected_result: OrcidRecord,
    error: type[Exception],
    error_message: str,
) -> None:
    if error:
        with pytest.raises(HTTPError, match=error_message):
            get_orcid_record_by_id(orcid_id)
    else:
        result = get_orcid_record_by_id(orcid_id)
        assert result == expected_result


jayne_carberry_result = OrcidRecord(
    orcid_identifier=OrcidIdentifier(
        path="0000-0003-4634-4047", uri="https://orcid.org/0000-0003-4634-4047"
    ),
    person=OrcidPerson(
        emails=OrcidEmails(email=[]),
        name=OrcidName(
            family_name=OrcidFamilyName(value="Carberry"),
            given_names=OrcidGivenNames(value="Jayne"),
            visibility="public",
        ),
    ),
)


@pytest.mark.usefixtures("mocked_orcid")
@pytest.mark.parametrize(
    ("search_string", "expected_result"),
    [
        ("Jayne Carberry", [jayne_carberry_result]),
        ("Multiple Carberry", [jayne_carberry_result] * 10),
    ],
    ids=["single result", "multiple results"],
)
def test_get_orcid_records_by_given_or_family_name(
    search_string: str, expected_result: list[OrcidRecord]
) -> None:
    response = search_records_by_name(given_and_family_names=search_string)
    assert response.total == len(expected_result)
    assert response.items == expected_result



from abc import abstractmethod
from collections.abc import Generator, Iterable

from mex.common.connector import BaseConnector
from mex.common.models import AnyExtractedModel, AnyMergedModel, AnyRuleSetResponse


class BaseSink(BaseConnector):
    """Base class to define the interface of sink instances."""

    def __init__(self) -> None:
        """Create a new sink."""

    def close(self) -> None:
        """Close the sink."""

    @abstractmethod
    def load(
        self,
        items: Iterable[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse],
    ) -> Generator[
        AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse, None, None
    ]:  # pragma: no cover
        """Load the given items to a destination and yield them."""
        ...



import json
import time
from collections.abc import Iterable
from types import NoneType
from typing import Annotated, Any, Literal

import pytest
from pydantic import computed_field
from pydantic.fields import FieldInfo

from mex.common.models import BaseModel
from mex.common.types import (
    MERGED_IDENTIFIER_CLASSES,
    Identifier,
    MergedPersonIdentifier,
)
from mex.common.utils import (
    GenericFieldInfo,
    any_contains_any,
    contains_any,
    contains_any_types,
    contains_only_types,
    deprecated,
    ensure_list,
    get_alias_lookup,
    get_all_fields,
    get_field_names_allowing_none,
    get_inner_types,
    get_list_field_names,
    group_fields_by_class_name,
    grouper,
    jitter_sleep,
)


@pytest.mark.parametrize(
    ("base", "tokens", "expected"),
    [
        ("foo", (), False),
        ("i have foo in it", ("foo", "bar", "batz"), True),
        ("i have none of the tokens ", ("foo", "bar", "batz"), False),
        ((1, 2, 3), (0, 2, 4), True),
    ],
    ids=["empty", "hit", "miss", "numbers"],
)
def test_contains_any(
    base: Any,  # noqa: ANN401
    tokens: Iterable[Any],
    expected: bool,  # noqa: FBT001
) -> None:
    assert contains_any(base, tokens) == expected


@pytest.mark.parametrize(
    ("base", "tokens", "expected"),
    [
        (["foo", None], (), False),
        (["i have foo in it", "i dont"], ("foo", "bar", "batz"), True),
        (["i have none of the tokens", "me neither"], ("foo", "bar", "batz"), False),
        (json.loads("[null, [1, 2], [], [3]]"), (0, 2, 4), True),
    ],
    ids=["empty", "hit", "miss", "numbers"],
)
def test_any_contains_any(
    base: Any,  # noqa: ANN401
    tokens: Iterable[Any],
    expected: bool,  # noqa: FBT001
) -> None:
    assert any_contains_any(base, tokens) == expected


@pytest.mark.parametrize(
    ("annotation", "types", "expected"),
    [
        (None, [str], False),
        (str, [str], True),
        (str, [Identifier], False),
        (Identifier, [str], False),
        (list[str | int | list[str]], [str, float], False),
        (list[str | int | list[str]], [int, str], True),
        (MergedPersonIdentifier | None, MERGED_IDENTIFIER_CLASSES, True),
    ],
    ids=[
        "static None",
        "simple str",
        "str vs identifier",
        "identifier vs str",
        "complex miss",
        "complex hit",
        "optional identifier",
    ],
)
def test_contains_only_types(
    annotation: Any,  # noqa: ANN401
    types: list[type],
    expected: bool,  # noqa: FBT001
) -> None:
    class DummyModel(BaseModel):
        attribute: annotation

    field_info = get_all_fields(DummyModel)["attribute"]
    assert contains_only_types(field_info, *types) == expected


@pytest.mark.parametrize(
    ("annotation", "types", "expected"),
    [
        (None, [str], False),
        (str, [str], True),
        (str, [Identifier], False),
        (Identifier, [str], False),
        (str, [int, str], True),
        (str, [int, float], False),
        (str | int, [str], True),
        (str | int, [int], True),
        (str | int, [float], False),
        (str | None, [str], True),
        (str | None, [NoneType], False),  # NoneType excluded by include_none=False
        (list[str], [str], True),
        (list[str], [list], False),  # list is unpacked by default
        (list[str | int], [str], True),
        (list[str | int], [int], True),
        (list[str | int], [float], False),
        (Annotated[str, "description"], [str], True),
        (Annotated[str | int, "description"], [str], True),
        (Annotated[str | int, "description"], [float], False),
        (Literal["value"], [Literal], True),
        (MergedPersonIdentifier | None, MERGED_IDENTIFIER_CLASSES, True),
        (MergedPersonIdentifier, MERGED_IDENTIFIER_CLASSES, True),
    ],
    ids=[
        "static None",
        "simple str match",
        "str vs identifier mismatch",
        "identifier vs str mismatch",
        "str in multiple types",
        "str not in multiple types",
        "union match first",
        "union match second",
        "union no match",
        "optional str match",
        "optional str excludes NoneType",
        "list of str",
        "list unpacked excludes list type",
        "nested union match first",
        "nested union match second",
        "nested union no match",
        "annotated str",
        "annotated union match",
        "annotated union no match",
        "literal type",
        "optional identifier match",
        "identifier match",
    ],
)
def test_contains_any_types(
    annotation: Any,  # noqa: ANN401
    types: list[type],
    expected: bool,  # noqa: FBT001
) -> None:
    field_info = GenericFieldInfo(alias=None, annotation=annotation, frozen=False)
    assert contains_any_types(field_info, *types) == expected


@pytest.mark.parametrize(
    ("annotation", "flags", "expected_types"),
    [
        (str, {}, [str]),
        (None, {}, [NoneType]),
        (None, {"include_none": False}, []),
        (str | None, {}, [str, NoneType]),
        (str | None, {"include_none": False}, [str]),
        (list[str] | None, {}, [str, NoneType]),
        (list[str | None], {}, [str, NoneType]),
        (list[int], {"unpack_list": False}, [list]),
        (list[str | int | list[str]], {}, [str, int, str]),
        (Annotated[str | int, FieldInfo(description="str or int")], {}, [str, int]),
        (Annotated[str | int, "This is a string or integer"], {}, [str, int]),
        (Literal["okay"] | None, {}, [Literal, NoneType]),
        (
            Literal["okay"] | None,
            {"unpack_literal": False},
            [Literal["okay"], NoneType],
        ),
    ],
    ids=[
        "string",
        "None allowing None",
        "None skipping None",
        "optional string allowing None",
        "optional string skipping None",
        "optional list of strings",
        "list of optional strings",
        "not unpacking list",
        "list nested in list",
        "annotated string or int with FieldInfo",
        "annotated string or int with plain text",
        "unpacking literal",
        "not unpacking literal",
    ],
)
def test_get_inner_types(
    annotation: Any,  # noqa: ANN401
    flags: dict[str, bool],
    expected_types: list[type],
) -> None:
    assert list(get_inner_types(annotation, **flags)) == expected_types


class Computer(BaseModel):
    ram: int = 16

    @computed_field  # type: ignore[prop-decorator]
    @property
    def cpus(self) -> int:  # pragma: no cover
        return 42


def test_get_all_fields_on_model_with_computed_field() -> None:
    assert get_all_fields(Computer) == {
        "cpus": GenericFieldInfo(alias=None, annotation=int, frozen=True),
        "ram": GenericFieldInfo(alias=None, annotation=int, frozen=False),
    }


class ComplexDummyModel(BaseModel):
    """Dummy Model with multiple attributes."""

    optional_str: str | None = None
    required_str: str = "default"
    optional_list: list[str] | None = None
    required_list: list[str] = []

    @computed_field(alias="computedInt")  # type: ignore[prop-decorator]
    @property
    def computed_int(self) -> int:  # pragma: no cover
        return 42


def test_get_alias_lookup() -> None:
    assert get_alias_lookup(ComplexDummyModel) == {
        "optional_str": "optional_str",
        "required_str": "required_str",
        "optional_list": "optional_list",
        "required_list": "required_list",
        "computedInt": "computed_int",
    }


def test_get_list_field_names() -> None:
    assert get_list_field_names(ComplexDummyModel) == [
        "optional_list",
        "required_list",
    ]


def test_get_field_names_allowing_none() -> None:
    assert get_field_names_allowing_none(ComplexDummyModel) == [
        "optional_str",
        "optional_list",
    ]


class DummyModel(BaseModel):
    number: int
    text: str


class PseudoModel(BaseModel):
    title: str


def test_group_fields_by_class_name() -> None:
    lookup: dict[str, type[BaseModel]] = {"Dummy": DummyModel, "Pseudo": PseudoModel}
    expected = {"Dummy": ["text"], "Pseudo": ["title"]}
    assert group_fields_by_class_name(lookup, lambda f: f.annotation is str) == expected


def test_grouper() -> None:
    groups = grouper(4, "ABCDEFGHIJ")
    joined = ["".join(c or "-" for c in g) for g in groups]
    assert joined == ["ABCD", "EFGH", "IJ--"]


def test_jitter_sleep() -> None:
    t0 = time.monotonic()
    jitter_sleep(0.1, 0.2)
    t1 = time.monotonic()
    time_slept = t1 - t0

    assert 0.1 < time_slept < 0.3  # giving 100ms of courtesy to pytest


@pytest.mark.parametrize(
    ("values", "expected"),
    [
        (None, []),
        ([], []),
        ([1, 2, 3], [1, 2, 3]),
        ("single_string", ["single_string"]),
        (42, [42]),
        (True, [True]),
        ({"key": "value"}, [{"key": "value"}]),
        ([None], [None]),
        (["nested", ["list"]], ["nested", ["list"]]),
    ],
    ids=[
        "none_to_empty_list",
        "empty_list_unchanged",
        "list_unchanged",
        "string_to_list",
        "number_to_list",
        "boolean_to_list",
        "dict_to_list",
        "list_with_none_unchanged",
        "nested_list_unchanged",
    ],
)
def test_ensure_list(values: Any, expected: list[Any]) -> None:  # noqa: ANN401
    assert ensure_list(values) == expected


def test_deprecated_warning_message() -> None:
    def cool_new_func() -> str:
        return "cool"

    old_func_bah = deprecated("old_func_bah", cool_new_func)

    with pytest.warns(DeprecationWarning, match="deprecated") as warning_info:
        old_func_bah()

    assert (
        str(warning_info[0].message)
        == "old_func_bah is deprecated, use cool_new_func instead"
    )



from typing import Any

import pytest

from mex.common.exceptions import MExError
from mex.common.merged.main import (
    _apply_lenient_fallback,
    _collect_additive_values,
    _collect_extracted_values,
    _collect_preventive_sources,
    _collect_subtractive_values,
    _create_merged_dict,
    _ensure_rule_set,
    _filter_usable_values,
    _get_merged_class,
    _pick_usable_values,
    create_merged_item,
)
from mex.common.merged.types import SourceAndValueList, SourceList, ValueList
from mex.common.models import (
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    ActivityRuleSetRequest,
    AdditivePerson,
    AdditiveResource,
    AnyExtractedModel,
    AnyMergedModel,
    AnyPreviewModel,
    AnyRuleSetRequest,
    ContactPointRuleSetRequest,
    ExtractedActivity,
    ExtractedContactPoint,
    ExtractedPerson,
    ExtractedResource,
    MergedPerson,
    PersonRuleSetRequest,
    PreventivePerson,
    PreventiveResource,
    PreviewPerson,
    ResourceRuleSetRequest,
    SubtractiveActivity,
    SubtractiveContactPoint,
    SubtractivePerson,
    SubtractiveResource,
)
from mex.common.testing import Joker
from mex.common.types import (
    AccessRestriction,
    AnyValidation,
    Identifier,
    MergedPrimarySourceIdentifier,
    Text,
    TextLanguage,
    Theme,
    Validation,
)


def test_collect_extracted_values() -> None:
    person = ExtractedPerson(
        fullName="Squidward, Dr. med.",
        email="squidward@ocean.clinic",
        hadPrimarySource=Identifier("squidwardSourceId"),
        identifierInPrimarySource="squidward",
    )

    result = _collect_extracted_values("email", [person])

    assert result == [(Identifier("squidwardSourceId"), "squidward@ocean.clinic")]


def test_collect_additive_values() -> None:
    rule_set = PersonRuleSetRequest(
        additive=AdditivePerson(givenName=["Bubbles", "Barnacle"])
    )

    result = _collect_additive_values("givenName", rule_set)

    assert result == [
        (MEX_PRIMARY_SOURCE_STABLE_TARGET_ID, "Bubbles"),
        (MEX_PRIMARY_SOURCE_STABLE_TARGET_ID, "Barnacle"),
    ]


def test_collect_subtractive_values() -> None:
    rule_set = PersonRuleSetRequest(
        subtractive=SubtractivePerson(
            email=["pineapple@fruit.ocean", "rock@ocean.floor"]
        )
    )

    result = _collect_subtractive_values("email", rule_set)

    assert result == ["pineapple@fruit.ocean", "rock@ocean.floor"]


def test_collect_preventive_sources() -> None:
    rule_set = PersonRuleSetRequest(
        preventive=PreventivePerson(email=[Identifier("preventPlankton")])
    )

    result = _collect_preventive_sources("email", rule_set)

    assert result == ["preventPlankton"]


@pytest.mark.parametrize(
    ("sources_values", "prevented", "subtracted", "expected"),
    [
        pytest.param([], [], [], [], id="empty_input"),
        pytest.param(
            [("src1", "value1"), ("src2", "value2")],
            [],
            [],
            ["value1", "value2"],
            id="no_filtering",
        ),
        pytest.param(
            [("src1", "value1"), ("src2", "value2")],
            ["src1"],
            [],
            ["value2"],
            id="prevent_source",
        ),
        pytest.param(
            [("src1", "value1"), ("src2", "value2")],
            [],
            ["value1"],
            ["value2"],
            id="subtract_value",
        ),
        pytest.param(
            [("src1", "value"), ("src2", "value")],
            [],
            [],
            ["value"],
            id="deduplicate",
        ),
    ],
)
def test_filter_usable_values(
    sources_values: SourceAndValueList,
    prevented: SourceList,
    subtracted: ValueList,
    expected: ValueList,
) -> None:
    result = _filter_usable_values(sources_values, prevented, subtracted)

    assert result == expected


def test_apply_lenient_fallback_empty() -> None:
    assert _apply_lenient_fallback((), (), []) == []


def test_apply_lenient_fallback() -> None:
    extracted = [(MergedPrimarySourceIdentifier("src1"), "extracted_value")]
    additive = [(MergedPrimarySourceIdentifier("src2"), "additive_value")]
    subtracted: ValueList = ["subtracted_value"]

    result = _apply_lenient_fallback(extracted, additive, subtracted)

    assert result == ["extracted_value"]


def test_pick_usable_values_strict() -> None:
    person = ExtractedPerson(
        fullName="Alice",
        email="alice@example.com",
        hadPrimarySource=Identifier.generate(seed=1),
        identifierInPrimarySource="alice",
    )
    rule_set = PersonRuleSetRequest()

    result = _pick_usable_values("email", [person], rule_set, Validation.STRICT)

    assert result == ["alice@example.com"]


def test_pick_usable_values_lenient() -> None:
    person = ExtractedPerson(
        fullName="Alice",
        hadPrimarySource=Identifier.generate(seed=1),
        identifierInPrimarySource="alice",
    )
    rule_set = PersonRuleSetRequest(
        subtractive=SubtractivePerson(email="blocked@example.com")
    )

    result = _pick_usable_values("email", [person], rule_set, Validation.LENIENT)

    assert result == ["blocked@example.com"]


def test_create_merged_dict_strict() -> None:
    person = ExtractedPerson(
        fullName="Alice",
        hadPrimarySource=Identifier.generate(seed=1),
        identifierInPrimarySource="alice",
    )
    rule_set = PersonRuleSetRequest()

    result = _create_merged_dict(["fullName"], [person], rule_set, Validation.STRICT)

    assert result == {"fullName": ["Alice"]}


@pytest.mark.parametrize(
    ("extracted_items", "rule_set", "validation", "expected"),
    [
        pytest.param(
            [],
            PersonRuleSetRequest(),
            Validation.STRICT,
            MergedPerson,
            id="strict_with_rules",
        ),
        pytest.param(
            [],
            PersonRuleSetRequest(),
            Validation.LENIENT,
            PreviewPerson,
            id="lenient_with_rules",
        ),
        pytest.param(
            [
                ExtractedPerson(
                    fullName="Test Person",
                    hadPrimarySource=Identifier.generate(seed=1),
                    identifierInPrimarySource="test",
                )
            ],
            None,
            Validation.STRICT,
            MergedPerson,
            id="strict_no_ruleset",
        ),
        pytest.param(
            [],
            None,
            Validation.STRICT,
            None,
            id="empty_inputs_returns_none",
        ),
    ],
)
def test_get_merged_class(
    extracted_items: list[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | None,
    validation: AnyValidation,
    expected: type[AnyMergedModel | AnyPreviewModel] | None,
) -> None:
    result = _get_merged_class(extracted_items, rule_set, validation)

    assert result == expected


def test_ensure_rule_set_creates_default() -> None:
    result = _ensure_rule_set(None, "Person")

    assert isinstance(result, PersonRuleSetRequest)


def test_ensure_rule_set_returns_existing() -> None:
    existing_rule_set = PersonRuleSetRequest()

    result = _ensure_rule_set(existing_rule_set, "Person")

    assert result is existing_rule_set


@pytest.mark.parametrize(
    ("extracted_items", "rule_set", "validation", "expected"),
    [
        pytest.param(
            [
                ExtractedResource(
                    identifierInPrimarySource="r1",
                    hadPrimarySource=Identifier.generate(seed=42),
                    accessRestriction=AccessRestriction["OPEN"],
                    contact=[Identifier.generate(seed=999)],
                    unitInCharge=[Identifier.generate(seed=999)],
                    theme=[Theme["PUBLIC_HEALTH"]],
                    title=[Text(value="Dummy resource")],
                )
            ],
            ResourceRuleSetRequest(
                additive=AdditiveResource(),
                subtractive=SubtractiveResource(),
                preventive=PreventiveResource(),
            ),
            Validation.STRICT,
            {
                "accessRestriction": "https://mex.rki.de/item/access-restriction-1",
                "contact": ["bFQoRhcVH5DIax"],
                "theme": ["https://mex.rki.de/item/theme-1"],
                "title": [{"value": "Dummy resource", "language": TextLanguage.EN}],
                "unitInCharge": ["bFQoRhcVH5DIax"],
                "entityType": "MergedResource",
                "identifier": "bFQoRhcVH5DHU6",
            },
            id="single extracted item",
        ),
        pytest.param(
            [
                ExtractedPerson(
                    fullName="Dr. Zoidberg",
                    affiliation=Identifier.generate(seed=99),
                    email="z@express.planet",
                    identifierInPrimarySource="drz",
                    hadPrimarySource=Identifier.generate(seed=9),
                ),
                ExtractedPerson(
                    fullName="Mr. Krabs",
                    email="manager@krusty.ocean",
                    affiliation=Identifier.generate(seed=101),
                    memberOf=[
                        Identifier.generate(seed=500),
                        Identifier.generate(seed=750),
                    ],
                    hadPrimarySource=Identifier.generate(seed=11),
                    identifierInPrimarySource="mrk",
                ),
            ],
            PersonRuleSetRequest(
                additive=AdditivePerson(
                    givenName=["Eugene", "Harold", "John"],
                    memberOf=[Identifier.generate(seed=500)],
                ),
                subtractive=SubtractivePerson(
                    email=["manager@krusty.ocean"],
                    givenName=["John"],
                ),
                preventive=PreventivePerson(
                    email=[Identifier.generate(seed=9)],
                    fullName=[Identifier.generate(seed=9)],
                ),
            ),
            Validation.STRICT,
            {
                "affiliation": [
                    Identifier.generate(seed=99),
                    Identifier.generate(seed=101),
                ],
                "fullName": ["Mr. Krabs"],
                "givenName": ["Eugene", "Harold"],
                "memberOf": [
                    Identifier.generate(seed=500),
                    Identifier.generate(seed=750),
                ],
                "identifier": Identifier.generate(seed=42),
                "entityType": "MergedPerson",
            },
            id="extracted items and rule set",
        ),
        pytest.param(
            [],
            PersonRuleSetRequest(
                additive=AdditivePerson(
                    givenName=["Eugene", "Harold", "John"],
                    memberOf=[Identifier.generate(seed=500)],
                ),
                subtractive=SubtractivePerson(
                    email=["manager@krusty.ocean"],
                    givenName=["John"],
                ),
                preventive=PreventivePerson(
                    email=[Identifier.generate(seed=9)],
                    fullName=[Identifier.generate(seed=9)],
                ),
            ),
            Validation.STRICT,
            {
                "givenName": ["Eugene", "Harold"],
                "memberOf": [
                    Identifier.generate(seed=500),
                ],
                "identifier": Identifier.generate(seed=42),
                "entityType": "MergedPerson",
            },
            id="only rule set",
        ),
        pytest.param(
            [
                ExtractedPerson(
                    fullName="Dr. Zoidberg",
                    affiliation=Identifier.generate(seed=99),
                    email="z@express.planet",
                    identifierInPrimarySource="drz",
                    hadPrimarySource=Identifier.generate(seed=9),
                ),
                ExtractedPerson(
                    fullName="Mr. Krabs",
                    email="manager@krusty.ocean",
                    affiliation=Identifier.generate(seed=101),
                    memberOf=[
                        Identifier.generate(seed=500),
                        Identifier.generate(seed=750),
                    ],
                    hadPrimarySource=Identifier.generate(seed=11),
                    identifierInPrimarySource="mrk",
                ),
            ],
            None,
            Validation.STRICT,
            {
                "affiliation": [
                    Identifier.generate(seed=99),
                    Identifier.generate(seed=101),
                ],
                "email": ["z@express.planet", "manager@krusty.ocean"],
                "fullName": ["Dr. Zoidberg", "Mr. Krabs"],
                "memberOf": [
                    Identifier.generate(seed=500),
                    Identifier.generate(seed=750),
                ],
                "identifier": Identifier.generate(seed=42),
                "entityType": "MergedPerson",
            },
            id="only extracted items",
        ),
        pytest.param(
            [],
            None,
            Validation.STRICT,
            "One of rule_set or extracted_items is required.",
            id="error if neither is supplied",
        ),
        pytest.param(
            [
                ExtractedContactPoint(
                    identifierInPrimarySource="krusty",
                    hadPrimarySource=Identifier.generate(seed=99),
                    email=["manager@krusty.ocean"],
                )
            ],
            ContactPointRuleSetRequest(
                subtractive=SubtractiveContactPoint(
                    email=["flipper@krusty.ocean", "manager@krusty.ocean"]
                )
            ),
            Validation.STRICT,
            "List should have at least 1 item after validation, not 0",
            id="merging raises cardinality error",
        ),
        pytest.param(
            [
                ExtractedActivity(
                    title=Text(value="Burger flipping"),
                    alternativeTitle=[],
                    contact=Identifier.generate(seed=97),
                    responsibleUnit=Identifier.generate(seed=98),
                    identifierInPrimarySource="BF",
                    hadPrimarySource=Identifier.generate(seed=99),
                ),
            ],
            ActivityRuleSetRequest(
                subtractive=SubtractiveActivity(
                    title=Text(value="Burger flipping"),
                    alternativeTitle=Text(value="Cash making"),
                ),
            ),
            Validation.LENIENT,
            {
                "title": [{"value": "Burger flipping", "language": None}],
                "alternativeTitle": [{"value": "Cash making", "language": None}],
                "contact": [Identifier.generate(seed=97)],
                "identifier": Identifier.generate(seed=42),
                "responsibleUnit": [Identifier.generate(seed=98)],
                "entityType": "PreviewActivity",
            },
            id="get preview of merged items",
        ),
        pytest.param(
            [
                ExtractedContactPoint(
                    identifierInPrimarySource="krusty",
                    hadPrimarySource=Identifier.generate(seed=99),
                    email=["manager@krusty.ocean"],
                )
            ],
            ContactPointRuleSetRequest(
                subtractive=SubtractiveContactPoint(
                    email=["flipper@krusty.ocean", "manager@krusty.ocean"]
                )
            ),
            Validation.LENIENT,
            {
                "identifier": Joker(),
                "entityType": "PreviewContactPoint",
                "email": ["manager@krusty.ocean"],
            },
            id="preview allows cardinality error",
        ),
        pytest.param(
            [
                ExtractedContactPoint(
                    identifierInPrimarySource="orders",
                    hadPrimarySource=Identifier.generate(seed=98),
                    email=["orders@krusty.ocean"],
                )
            ],
            ContactPointRuleSetRequest(
                subtractive=SubtractiveContactPoint(
                    email=["orders@krusty.ocean"],
                )
            ),
            Validation.IGNORE,
            None,
            id="ignore mode validation returns none",
        ),
        pytest.param(
            [
                ExtractedContactPoint(
                    email=[f"{i}@contact-point.com"],
                    hadPrimarySource=Identifier.generate(seed=1),
                    identifierInPrimarySource=f"{i}",
                )
                for i in range(20)
            ],
            ContactPointRuleSetRequest(),
            Validation.IGNORE,
            {
                "email": [
                    "4@contact-point.com",
                    "15@contact-point.com",
                    "1@contact-point.com",
                    "12@contact-point.com",
                    "19@contact-point.com",
                    "11@contact-point.com",
                    "6@contact-point.com",
                    "16@contact-point.com",
                    "10@contact-point.com",
                    "17@contact-point.com",
                    "18@contact-point.com",
                    "14@contact-point.com",
                    "8@contact-point.com",
                    "9@contact-point.com",
                    "7@contact-point.com",
                    "2@contact-point.com",
                    "0@contact-point.com",
                    "3@contact-point.com",
                    "5@contact-point.com",
                    "13@contact-point.com",
                ],
                "entityType": "MergedContactPoint",
                "identifier": Joker(),
            },
            id="stable order by identifier",
        ),
    ],
)
def test_create_merged_item(
    extracted_items: list[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | None,
    validation: AnyValidation,
    expected: dict[str, Any] | str | None,
) -> None:
    try:
        merged_item = create_merged_item(
            Identifier.generate(seed=42),
            extracted_items,
            rule_set,
            validation,
        )
    except MExError as error:
        assert str(expected) in f"{error}: {error.__cause__}"  # noqa: PT017
    else:
        if merged_item is None:
            assert expected is None
        else:
            clean_dict = {k: v for k, v in merged_item.model_dump().items() if v}
            assert clean_dict == expected



from unittest.mock import MagicMock, Mock

import pytest
from pytest import MonkeyPatch

from mex.common.backend_api.connector import BackendApiConnector
from mex.common.models import ExtractedPerson, MergedPerson
from mex.common.sinks.backend_api import BackendApiSink


def test_sink_load_mocked(
    extracted_person: ExtractedPerson, monkeypatch: MonkeyPatch
) -> None:
    def __init__(self: BackendApiConnector) -> None:
        self.session = MagicMock()

    monkeypatch.setattr(BackendApiConnector, "__init__", __init__)

    ingest = Mock(return_value=[extracted_person])
    monkeypatch.setattr(BackendApiConnector, "ingest", ingest)

    sink = BackendApiSink.get()
    models_or_rule_sets = list(sink.load([extracted_person]))
    assert models_or_rule_sets == [extracted_person]
    ingest.assert_called_once_with([extracted_person], timeout=(5, 90))


def test_sink_load_merged_error(
    merged_person: MergedPerson, monkeypatch: MonkeyPatch
) -> None:
    def __init__(self: BackendApiConnector) -> None:
        self.session = MagicMock()

    monkeypatch.setattr(BackendApiConnector, "__init__", __init__)

    sink = BackendApiSink.get()
    with pytest.raises(
        NotImplementedError,
        match="Backend can only ingest extracted models and rule-set responses",
    ):
        list(sink.load([merged_person]))


def test_sink_is_supported(
    extracted_person: ExtractedPerson,
    merged_person: MergedPerson,
) -> None:
    assert BackendApiSink.is_supported([extracted_person, extracted_person]) is True
    assert BackendApiSink.is_supported([extracted_person, merged_person]) is False



from mex.common.models import ExtractedPrimarySource
from mex.common.primary_source.helpers import (
    get_extracted_primary_source_by_name,
)


def test_get_extracted_primary_source_by_name() -> None:
    string_wiki = "wikidata"
    string_nonsense = "this should give no match"

    # primary source found
    extracted_primary_source = get_extracted_primary_source_by_name(string_wiki)

    assert isinstance(extracted_primary_source, ExtractedPrimarySource)
    assert extracted_primary_source.identifierInPrimarySource == "wikidata"

    assert get_extracted_primary_source_by_name(string_nonsense) is None



from typing import Annotated, ClassVar, Literal

from pydantic import AfterValidator, Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ActivityType,
    ExtractedActivityIdentifier,
    Identifier,
    Link,
    MergedActivityIdentifier,
    MergedBibliographicResourceIdentifier,
    MergedContactPointIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
    Theme,
    Year,
    YearMonth,
    YearMonthDay,
)

AnyExternalAssociateIdentifier = Annotated[
    MergedOrganizationIdentifier | MergedPersonIdentifier,
    AfterValidator(Identifier),
]
AnyContactIdentifier = Annotated[
    MergedOrganizationalUnitIdentifier
    | MergedPersonIdentifier
    | MergedContactPointIdentifier,
    AfterValidator(Identifier),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Activity"], Field(frozen=True)]] = "Activity"


class _OptionalLists(_Stem):
    abstract: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/abstract"]}),
    ] = []
    activityType: Annotated[
        list[ActivityType],
        Field(json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]}),
    ] = []
    alternativeTitle: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]}),
    ] = []
    documentation: Annotated[
        list[Link],
        Field(
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/isReferencedBy"]
            }
        ),
    ] = []
    end: Annotated[
        list[YearMonthDay | YearMonth | Year],
        Field(json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P582"]}),
    ] = []
    externalAssociate: Annotated[
        list[AnyExternalAssociateIdentifier],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]}),
    ] = []
    funderOrCommissioner: Annotated[
        list[MergedOrganizationIdentifier],
        Field(json_schema_extra={"sameAs": "http://www.wikidata.org/entity/P8324"}),
    ] = []
    fundingProgram: list[str] = []
    involvedPerson: Annotated[
        list[MergedPersonIdentifier],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]}),
    ] = []
    involvedUnit: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]}),
    ] = []
    isPartOfActivity: Annotated[
        list[MergedActivityIdentifier],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://purl.org/dc/terms/isPartOf",
                    "http://www.cidoc-crm.org/cidoc-crm/P9i_forms_part_of",
                ]
            }
        ),
    ] = []
    publication: Annotated[
        list[MergedBibliographicResourceIdentifier],
        Field(
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/isReferencedBy"]
            }
        ),
    ] = []
    shortName: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P1813"]}),
    ] = []
    start: Annotated[
        list[YearMonthDay | YearMonth | Year],
        Field(json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P580"]}),
    ] = []
    succeeds: Annotated[
        list[MergedActivityIdentifier],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.cidoc-crm.org/cidoc-crm/P173_start_before_or_with_the_end_of"
                ]
            }
        ),
    ] = []
    theme: Annotated[
        list[Theme],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#theme"]}),
    ] = []
    website: Annotated[
        list[Link],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.wikidata.org/entity/P856",
                    "http://xmlns.com/foaf/0.1/homepage",
                ]
            }
        ),
    ] = []


class _RequiredLists(_Stem):
    contact: Annotated[
        list[AnyContactIdentifier],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#contactPoint"]},
        ),
    ]
    responsibleUnit: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": "http.//dcat-ap.de/def/dcatde/maintainer"},
        ),
    ]
    title: Annotated[
        list[Text],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ]


class _SparseLists(_Stem):
    contact: list[AnyContactIdentifier] = []
    responsibleUnit: list[MergedOrganizationalUnitIdentifier] = []
    title: list[Text] = []


class BaseActivity(
    _OptionalLists,
    _RequiredLists,
    json_schema_extra={
        "description": (
            "An activity carried out by RKI. This may be a research activity, such as "
            "a funded project, or a task that RKI performs under federal law.  "
            "Activities provide useful context information for resources."
        ),
        "sameAs": [
            "http://www.cidoc-crm.org/cidoc-crm/E7_Activity",
            "http://purl.org/dc/terms/Activity",
            "http://www.w3.org/ns/prov#Activity",
        ],
        "title": "Activity",
    },
):
    """All fields for a valid activity except for provenance."""


class ExtractedActivity(BaseActivity, ExtractedData):
    """An automatically extracted metadata set describing an activity."""

    entityType: Annotated[
        Literal["ExtractedActivity"], Field(alias="$type", frozen=True)
    ] = "ExtractedActivity"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedActivityIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedActivityIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedActivityIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedActivityIdentifier)


class MergedActivity(BaseActivity, MergedItem):
    """The result of merging all extracted items and rules for an activity."""

    entityType: Annotated[
        Literal["MergedActivity"], Field(alias="$type", frozen=True)
    ] = "MergedActivity"
    identifier: Annotated[MergedActivityIdentifier, Field(frozen=True)]


class PreviewActivity(_OptionalLists, _SparseLists, PreviewItem):
    """Preview for merging all extracted items and rules for an activity."""

    entityType: Annotated[
        Literal["PreviewActivity"], Field(alias="$type", frozen=True)
    ] = "PreviewActivity"
    identifier: Annotated[MergedActivityIdentifier, Field(frozen=True)]


class AdditiveActivity(_OptionalLists, _SparseLists, AdditiveRule):
    """Rule to add values to merged activity items."""

    entityType: Annotated[
        Literal["AdditiveActivity"], Field(alias="$type", frozen=True)
    ] = "AdditiveActivity"


class SubtractiveActivity(_OptionalLists, _SparseLists, SubtractiveRule):
    """Rule to subtract values from merged activity items."""

    entityType: Annotated[
        Literal["SubtractiveActivity"], Field(alias="$type", frozen=True)
    ] = "SubtractiveActivity"


class PreventiveActivity(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged activity items."""

    entityType: Annotated[
        Literal["PreventiveActivity"], Field(alias="$type", frozen=True)
    ] = "PreventiveActivity"
    abstract: list[MergedPrimarySourceIdentifier] = []
    activityType: list[MergedPrimarySourceIdentifier] = []
    alternativeTitle: list[MergedPrimarySourceIdentifier] = []
    contact: list[MergedPrimarySourceIdentifier] = []
    documentation: list[MergedPrimarySourceIdentifier] = []
    end: list[MergedPrimarySourceIdentifier] = []
    externalAssociate: list[MergedPrimarySourceIdentifier] = []
    funderOrCommissioner: list[MergedPrimarySourceIdentifier] = []
    fundingProgram: list[MergedPrimarySourceIdentifier] = []
    involvedPerson: list[MergedPrimarySourceIdentifier] = []
    involvedUnit: list[MergedPrimarySourceIdentifier] = []
    isPartOfActivity: list[MergedPrimarySourceIdentifier] = []
    publication: list[MergedPrimarySourceIdentifier] = []
    responsibleUnit: list[MergedPrimarySourceIdentifier] = []
    shortName: list[MergedPrimarySourceIdentifier] = []
    start: list[MergedPrimarySourceIdentifier] = []
    succeeds: list[MergedPrimarySourceIdentifier] = []
    theme: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []
    website: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditiveActivity = AdditiveActivity()
    subtractive: SubtractiveActivity = SubtractiveActivity()
    preventive: PreventiveActivity = PreventiveActivity()


class ActivityRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update an activity item."""

    entityType: Annotated[
        Literal["ActivityRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "ActivityRuleSetRequest"


class ActivityRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve an activity item."""

    entityType: Annotated[
        Literal["ActivityRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "ActivityRuleSetResponse"
    stableTargetId: MergedActivityIdentifier


class ActivityMapping(_Stem, BaseMapping):
    """Mapping for describing an activity transformation."""

    entityType: Annotated[
        Literal["ActivityMapping"], Field(alias="$type", frozen=True)
    ] = "ActivityMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    contact: Annotated[
        list[MappingField[list[AnyContactIdentifier]]], Field(min_length=1)
    ]
    responsibleUnit: Annotated[
        list[MappingField[list[MergedOrganizationalUnitIdentifier]]],
        Field(min_length=1),
    ]
    title: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    abstract: list[MappingField[list[Text]]] = []
    activityType: list[MappingField[list[ActivityType]]] = []
    alternativeTitle: list[MappingField[list[Text]]] = []
    documentation: list[MappingField[list[Link]]] = []
    end: list[MappingField[list[YearMonthDay | YearMonth | Year]]] = []
    externalAssociate: list[MappingField[list[AnyExternalAssociateIdentifier]]] = []
    funderOrCommissioner: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    fundingProgram: list[MappingField[list[str]]] = []
    involvedPerson: list[MappingField[list[MergedPersonIdentifier]]] = []
    involvedUnit: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []
    isPartOfActivity: list[MappingField[list[MergedActivityIdentifier]]] = []
    publication: list[MappingField[list[MergedBibliographicResourceIdentifier]]] = []
    shortName: list[MappingField[list[Text]]] = []
    start: list[MappingField[list[YearMonthDay | YearMonth | Year]]] = []
    succeeds: list[MappingField[list[MergedActivityIdentifier]]] = []
    theme: list[MappingField[list[Theme]]] = []
    website: list[MappingField[list[Link]]] = []


class ActivityFilter(_Stem, BaseFilter):
    """Class for defining filter rules for activity items."""

    entityType: Annotated[
        Literal["ActivityFilter"], Field(alias="$type", frozen=True)
    ] = "ActivityFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from mex.common.models import ExtractedOrganizationalUnit
from mex.common.organigram.extract import (
    extract_organigram_units,
    get_extracted_unit_by_synonyms,
    get_unit_merged_ids_by_emails,
    get_unit_merged_ids_by_synonyms,
)
from mex.common.organigram.models import OrganigramUnit
from mex.common.types import Text


def test_extract_organigram_units(
    child_unit: OrganigramUnit, parent_unit: OrganigramUnit
) -> None:
    units = list(extract_organigram_units())
    assert units == [child_unit, parent_unit]


def test_get_extracted_unit_by_synonyms(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    mapping = get_extracted_unit_by_synonyms(
        [extracted_child_unit, extracted_parent_unit]
    )

    assert (
        mapping["Abteilung"][0].stableTargetId == extracted_parent_unit.stableTargetId
    )
    assert mapping["C1"][0].stableTargetId == extracted_child_unit.stableTargetId


def test_get_unit_merged_ids_by_synonyms(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    mapping = get_unit_merged_ids_by_synonyms(
        [extracted_child_unit, extracted_parent_unit]
    )
    child_id = extracted_child_unit.stableTargetId
    parent_id = extracted_parent_unit.stableTargetId
    assert mapping == {
        "Abteilung": [parent_id],
        "C1": [child_id],
        "C1 Sub-Unit": [child_id],
        "C1 Unterabteilung": [child_id],
        "C1: Sub Unit": [child_id],
        "CHLD": [child_id],
        "CHLD Unterabteilung": [child_id],
        "child-unit": [child_id],
        "Department": [parent_id],
        "PARENT Dept.": [parent_id],
        "PRNT": [parent_id],
        "PRNT Abteilung": [parent_id],
        "parent-unit": [parent_id],
    }


def test_get_extracted_unit_by_synonyms_multiple_units_same_label(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    extracted_child_unit.name.append(Text(value="SHARED Dept."))
    extracted_parent_unit.name.append(Text(value="SHARED Dept."))

    mapping = get_extracted_unit_by_synonyms(
        [extracted_child_unit, extracted_parent_unit]
    )

    assert "SHARED Dept." in mapping
    assert len(mapping["SHARED Dept."]) == 2
    assert extracted_child_unit in mapping["SHARED Dept."]
    assert extracted_parent_unit in mapping["SHARED Dept."]


def test_get_unit_merged_ids_by_synonyms_multiple_units_same_label(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    extracted_child_unit.name.append(Text(value="SHARED Dept."))
    extracted_parent_unit.name.append(Text(value="SHARED Dept."))

    mapping = get_unit_merged_ids_by_synonyms(
        [extracted_child_unit, extracted_parent_unit]
    )

    assert "SHARED Dept." in mapping
    assert len(mapping["SHARED Dept."]) == 2
    assert extracted_child_unit.stableTargetId in mapping["SHARED Dept."]
    assert extracted_parent_unit.stableTargetId in mapping["SHARED Dept."]


def test_get_unit_merged_ids_by_emails(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    mapping = get_unit_merged_ids_by_emails(
        [extracted_child_unit, extracted_parent_unit]
    )
    assert mapping == {
        "parent@example.com": [extracted_parent_unit.stableTargetId],
        "pu@example.com": [extracted_parent_unit.stableTargetId],
        # child unit has no emails
    }


def test_get_unit_merged_ids_by_emails_multiple_units_same_email(
    extracted_child_unit: ExtractedOrganizationalUnit,
    extracted_parent_unit: ExtractedOrganizationalUnit,
) -> None:
    """Test that multiple units with the same email are all included in the list."""
    extracted_child_unit.email = ["shared@example.com"]
    extracted_parent_unit.email = ["shared@example.com", "parent@example.com"]

    mapping = get_unit_merged_ids_by_emails(
        [extracted_child_unit, extracted_parent_unit]
    )

    assert "shared@example.com" in mapping
    assert len(mapping["shared@example.com"]) == 2
    assert extracted_child_unit.stableTargetId in mapping["shared@example.com"]
    assert extracted_parent_unit.stableTargetId in mapping["shared@example.com"]

    assert mapping["parent@example.com"] == [extracted_parent_unit.stableTargetId]



"""Helper extractor to extract organigram data.

Common use cases:
- extract organizational units of your organization (e.g. president, departments, units)

Configuration
-------------

The extractor reads data from a json file, whose path is set with the settings parameter
`organigram_path` (see `mex.common.settings` for further info).

Extracting data
---------------

The module `organigram.extract` contains functions for data extraction.

Transforming data
-----------------

The module `organigram.transform` contains functions for data transformation.

Use the `stableTargetId` attribute of the transformed objects to set attributes
requiring an `MergedOrganizationalUnitIdentifier`.

Convenience Functions
---------------------

The module `organigram.extract` holds convenience functions, e.g. for building a mapping
from email addresses or synonyms to `ExtractedOrganizationalUnit`.

JSON file format
----------------

See example file in `assets/raw-data/organigram/organizational_units.json`.
For mandatory / optional attributes, see model in `mex/common/organigram/models.py`.
"""



from datetime import date
from io import StringIO

from pytest import LogCaptureFixture

from mex.common.extract import get_dtypes_for_model, parse_csv
from mex.common.models import BaseModel


class DummyModel(BaseModel):
    bool_: bool
    str_: str
    date_: date
    float_: float
    int_: int


def test_get_dtypes_for_model() -> None:
    assert get_dtypes_for_model(DummyModel) == {
        "bool_": "bool",
        "str_": "string",
        "date_": "string",
        "float_": "Float64",
        "int_": "Int64",
    }


def test_parse_csv(caplog: LogCaptureFixture) -> None:
    buffer = StringIO(
        """
bool_,str_,date_,float_,int_
true,"good row",2000-01-01,2.718,42
false,"bad row",,,
false,"another bad row",2000-01-01,,""
true,"another good row",2001-02-03,3.14,42
        """.strip()
    )

    parsed_models = list(parse_csv(buffer, DummyModel, chunksize=2))
    assert len(parsed_models) == 2
    assert parsed_models[0].str_ == "good row"
    assert parsed_models[1].str_ == "another good row"

    records = caplog.records

    assert (
        len(records) == 7
    )  # 2 info logs for chunks, 1+3 error logs and 1 info log for amount processed

    assert records[0].levelname == "INFO"
    assert "parse_csv - DummyModel chunk 0 - OK" in records[0].message
    assert records[1].levelname == "INFO"
    assert "parse_csv - DummyModel chunk 1 - OK" in records[1].message

    assert records[2].levelname == "ERROR"
    assert "Summarizing errors for remaining rows" in records[2].message
    assert records[3].levelname == "ERROR"
    assert "Error type 'date_type': 1 occurrences" in records[3].message
    assert records[4].levelname == "ERROR"
    assert "Error type 'float_type': 2 occurrences" in records[4].message
    assert records[5].levelname == "ERROR"
    assert "Error type 'int_type': 2 occurrences" in records[5].message

    assert records[6].levelname == "INFO"
    assert "Successfully processed 2 items." in records[6].message


def test_parse_csv_batch_summary(caplog: LogCaptureFixture) -> None:
    buffer = StringIO(
        """
bool_,str_,date_,float_,int_
true,"good",2000-01-01,1.0,1
false,"bad1",,,
false,"bad2",,,
true,"good",2000-01-01,1.0,1
false,"bad3",,,
        """.strip()
    )

    parsed_models = list(parse_csv(buffer, DummyModel, summary_batch_size=3))
    assert len(parsed_models) == 2
    assert parsed_models[0].str_ == "good"
    assert parsed_models[1].str_ == "good"

    records = caplog.records

    assert (
        len(records) == 10
    )  # 1 info log for chunk, 1+3+1+3 error logs for 2 batches and 1 info log for amount processed

    assert records[0].levelname == "INFO"
    assert "parse_csv - DummyModel chunk 0 - OK" in records[0].message

    assert records[1].levelname == "ERROR"
    assert "Summarizing errors for batch with rows 1 to 3" in records[1].message
    assert records[2].levelname == "ERROR"
    assert "Error type 'date_type': 2 occurrences" in records[2].message
    assert records[3].levelname == "ERROR"
    assert "Error type 'float_type': 2 occurrences" in records[3].message
    assert records[4].levelname == "ERROR"
    assert "Error type 'int_type': 2 occurrences" in records[4].message

    assert records[5].levelname == "ERROR"
    assert "Summarizing errors for remaining rows" in records[5].message
    assert records[6].levelname == "ERROR"
    assert "Error type 'date_type': 1 occurrences" in records[6].message
    assert records[7].levelname == "ERROR"
    assert "Error type 'float_type': 1 occurrences" in records[7].message
    assert records[8].levelname == "ERROR"
    assert "Error type 'int_type': 1 occurrences" in records[8].message

    assert records[9].levelname == "INFO"
    assert "Successfully processed 2 items." in records[9].message



import platform
import re
from pathlib import Path

import pytest
from pydantic import ValidationError

from mex.common.models import BaseModel
from mex.common.settings import SETTINGS_STORE, BaseSettings
from mex.common.types import AssetsPath, WorkPath


def test_debug_setting() -> None:
    # Test settings can be instantiated as basic sanity check
    settings = BaseSettings(debug=True)

    assert settings.debug is True


def test_settings_text() -> None:
    # Test settings can be converted to a legible multi-line paragraph
    settings = BaseSettings.get()
    text = settings.text()

    assert len(text.splitlines()) == len(BaseSettings.model_fields)
    assert re.search(r"debug\s+False", text)
    assert re.search(r"backend_api_key\s+\*+", text)  # masked secret


class FooSettings(BaseSettings):
    """Dummy settings subclass for testing."""

    foo: str = "foo"


class BarSettings(BaseSettings):
    """Second dummy settings subclass for testing."""

    bar: str = "bar"


class BazSettings(BaseSettings):
    """Third dummy settings subclass for testing."""

    foo: BaseSettings = FooSettings()


def test_settings_getting_caches_singleton() -> None:
    # clear cache
    SETTINGS_STORE.reset()

    # first get
    settings = FooSettings.get()
    cached_settings = SETTINGS_STORE.load(FooSettings)
    assert settings is cached_settings

    # repeated get
    settings_fetched_again = FooSettings.get()
    assert settings_fetched_again is settings


@pytest.mark.integration
def test_parse_env_file() -> None:
    settings = BaseSettings.get()
    # "work_dir" and "assets_dir" are always set, assert that more than these two are
    # set. This indicates an .env file was found and at least one setting was parsed.
    assert settings.model_fields_set != {"work_dir", "assets_dir"}


class SubModel(BaseModel):
    sub_model_path: WorkPath


class DummySettings(BaseSettings):
    non_path: str
    abs_work_path: WorkPath
    rel_work_path: WorkPath
    assets_path: AssetsPath
    sub_model: SubModel


def test_resolve_paths() -> None:
    if platform.system() == "Windows":  # pragma: no cover
        absolute = WorkPath(r"C:\absolute\path")
    else:  # pragma: no cover
        absolute = WorkPath("/absolute/path")

    relative = WorkPath(Path("relative", "path"))

    settings = DummySettings(
        non_path="blablabla",
        abs_work_path=absolute,
        rel_work_path=relative,
        assets_path=AssetsPath(relative),
        assets_dir=Path(absolute / "assets_dir"),
        work_dir=Path(absolute / "work_dir"),
        sub_model=SubModel(sub_model_path=relative),
    )

    assert settings.non_path == "blablabla"
    assert settings.abs_work_path == absolute
    assert settings.rel_work_path == WorkPath(settings.work_dir / relative)
    assert settings.assets_path == AssetsPath(absolute / "assets_dir" / relative)
    assert settings.sub_model.sub_model_path == WorkPath(settings.work_dir / relative)


def test_no_settings_as_attributes() -> None:
    # regular settings instantiation passes
    FooSettings()

    # setting with attribute that inherits from BaseSettings fails
    with pytest.raises(ValidationError):
        BazSettings()



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedOrganizationalUnitIdentifier,
    Link,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
)

EmailStr = Annotated[
    str,
    Field(
        examples=["info@rki.de"],
        pattern="^[^@ \\t\\r\\n]+@[^@ \\t\\r\\n]+\\.[^@ \\t\\r\\n]+$",
        json_schema_extra={"format": "email"},
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["OrganizationalUnit"], Field(frozen=True)]] = (
        "OrganizationalUnit"
    )


class _OptionalLists(_Stem):
    alternativeName: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]}),
    ] = []
    email: Annotated[
        list[EmailStr],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/2006/vcard/ns#hasEmail",
                    "https://schema.org/email",
                ]
            }
        ),
    ] = []
    shortName: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P1813"]}),
    ] = []
    unitOf: Annotated[
        list[MergedOrganizationIdentifier],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/org#unitOf"]}),
    ] = []
    website: Annotated[
        list[Link],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.wikidata.org/entity/P856",
                    "http://www.w3.org/2006/vcard/ns#hasUrl",
                    "http://xmlns.com/foaf/0.1/homepage",
                ]
            }
        ),
    ] = []


class _RequiredLists(_Stem):
    name: Annotated[
        list[Text],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": "http://xmlns.com/foaf/0.1/name"},
        ),
    ]


class _SparseLists(_Stem):
    name: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": "http://xmlns.com/foaf/0.1/name"}),
    ] = []


class _OptionalValues(_Stem):
    parentUnit: MergedOrganizationalUnitIdentifier | None = None


class _VariadicValues(_Stem):
    parentUnit: list[MergedOrganizationalUnitIdentifier] = []


class BaseOrganizationalUnit(
    _OptionalLists,
    _RequiredLists,
    _OptionalValues,
    json_schema_extra={
        "description": (
            "An Organization such as a department or support unit which is part of "
            "some larger Organization and only has full recognition within the context "
            "of that Organization. In particular the unit would not be regarded as a "
            "legal entity in its own right."
        ),
        "sameAs": [
            "http://www.w3.org/ns/org#OrganizationalUnit",
            "http://www.w3.org/2006/vcard/ns#Group",
            "http://www.cidoc-crm.org/cidoc-crm/E_74_Group",
        ],
        "title": "Organizational Unit",
    },
):
    """All fields for a valid organizational unit except for provenance."""


class ExtractedOrganizationalUnit(BaseOrganizationalUnit, ExtractedData):
    """An automatically extracted metadata set describing an organizational unit."""

    entityType: Annotated[
        Literal["ExtractedOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "ExtractedOrganizationalUnit"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedOrganizationalUnitIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedOrganizationalUnitIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedOrganizationalUnitIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedOrganizationalUnitIdentifier)


class MergedOrganizationalUnit(BaseOrganizationalUnit, MergedItem):
    """The result of merging all extracted items and rules for an organizational unit."""  # noqa: E501

    entityType: Annotated[
        Literal["MergedOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "MergedOrganizationalUnit"
    identifier: Annotated[MergedOrganizationalUnitIdentifier, Field(frozen=True)]


class PreviewOrganizationalUnit(
    _OptionalLists, _SparseLists, _OptionalValues, PreviewItem
):
    """Preview for merging all extracted items and rules for an organizational unit."""

    entityType: Annotated[
        Literal["PreviewOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "PreviewOrganizationalUnit"
    identifier: Annotated[MergedOrganizationalUnitIdentifier, Field(frozen=True)]


class AdditiveOrganizationalUnit(
    _OptionalLists, _SparseLists, _OptionalValues, AdditiveRule
):
    """Rule to add values to merged organizational units."""

    entityType: Annotated[
        Literal["AdditiveOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "AdditiveOrganizationalUnit"


class SubtractiveOrganizationalUnit(
    _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
):
    """Rule to subtract values from merged organizational units."""

    entityType: Annotated[
        Literal["SubtractiveOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "SubtractiveOrganizationalUnit"


class PreventiveOrganizationalUnit(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged organizational units."""

    entityType: Annotated[
        Literal["PreventiveOrganizationalUnit"], Field(alias="$type", frozen=True)
    ] = "PreventiveOrganizationalUnit"
    alternativeName: list[MergedPrimarySourceIdentifier] = []
    email: list[MergedPrimarySourceIdentifier] = []
    name: list[MergedPrimarySourceIdentifier] = []
    parentUnit: list[MergedPrimarySourceIdentifier] = []
    shortName: list[MergedPrimarySourceIdentifier] = []
    unitOf: list[MergedPrimarySourceIdentifier] = []
    website: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditiveOrganizationalUnit = AdditiveOrganizationalUnit()
    subtractive: SubtractiveOrganizationalUnit = SubtractiveOrganizationalUnit()
    preventive: PreventiveOrganizationalUnit = PreventiveOrganizationalUnit()


class OrganizationalUnitRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update an organizational unit item."""

    entityType: Annotated[
        Literal["OrganizationalUnitRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "OrganizationalUnitRuleSetRequest"


class OrganizationalUnitRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve an organizational unit item."""

    entityType: Annotated[
        Literal["OrganizationalUnitRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "OrganizationalUnitRuleSetResponse"
    stableTargetId: MergedOrganizationalUnitIdentifier


class OrganizationalUnitMapping(_Stem, BaseMapping):
    """Mapping for describing an organizational unit transformation."""

    entityType: Annotated[
        Literal["OrganizationalUnitMapping"], Field(alias="$type", frozen=True)
    ] = "OrganizationalUnitMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    parentUnit: list[MappingField[MergedOrganizationalUnitIdentifier | None]] = []
    name: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    alternativeName: list[MappingField[list[Text]]] = []
    email: list[MappingField[list[EmailStr]]] = []
    shortName: list[MappingField[list[Text]]] = []
    unitOf: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    website: list[MappingField[list[Link]]] = []


class OrganizationalUnitFilter(_Stem, BaseFilter):
    """Class for defining filter rules for organizational unit items."""

    entityType: Annotated[
        Literal["OrganizationalUnitFilter"], Field(alias="$type", frozen=True)
    ] = "OrganizationalUnitFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []
